# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
import struct
from functools import reduce

from .memory_element import MemoryElement

logger = logging.getLogger(__name__)


EEPROM_TOKEN = b'0xBC'


class I2CElement(MemoryElement):

    def __init__(self, id, type, size, mem_handler):
        super(I2CElement, self).__init__(id=id, type=type, size=size,
                                         mem_handler=mem_handler)
        self._update_finished_cb = None
        self._write_finished_cb = None
        self.elements = {}
        self.valid = False

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            if addr == 0:
                done = False
                # Check for header
                if data[0:4] == EEPROM_TOKEN:
                    logger.debug('Got new data: {}'.format(data))
                    [self.elements['version'],
                     self.elements['radio_channel'],
                     self.elements['radio_speed'],
                     self.elements['pitch_trim'],
                     self.elements['roll_trim']] = struct.unpack('<BBBff',
                                                                 data[4:15])
                    if self.elements['version'] == 0:
                        done = True
                    elif self.elements['version'] == 1:
                        self.datav0 = data
                        self.mem_handler.read(self, 16, 5)
                else:
                    self.valid = False
                    if self._update_finished_cb:
                        self._update_finished_cb(self)
                        self._update_finished_cb = None

            if addr == 16:
                [radio_address_upper, radio_address_lower] = struct.unpack(
                    '<BI', self.datav0[15:16] + data[0:4])
                self.elements['radio_address'] = int(
                    radio_address_upper) << 32 | radio_address_lower

                logger.debug(self.elements)
                data = self.datav0 + data
                done = True

            if done:
                if self._checksum256(data[:len(data) - 1]) == \
                        data[len(data) - 1]:
                    self.valid = True
                if self._update_finished_cb:
                    self._update_finished_cb(self)
                    self._update_finished_cb = None

    def _checksum256(self, st):
        return reduce(lambda x, y: x + y, list(st)) % 256

    def write_data(self, write_finished_cb):
        image = bytearray()
        if self.elements['version'] == 0:
            data = (
                0x00, self.elements['radio_channel'],
                self.elements['radio_speed'],
                self.elements['pitch_trim'], self.elements['roll_trim'])
            image += struct.pack('<BBBff', *data)
        elif self.elements['version'] == 1:
            data = (
                0x01, self.elements['radio_channel'],
                self.elements['radio_speed'],
                self.elements['pitch_trim'], self.elements['roll_trim'],
                self.elements['radio_address'] >> 32,
                self.elements['radio_address'] & 0xFFFFFFFF)
            image += struct.pack('<BBBffBI', *data)
        # Adding some magic:
        image = EEPROM_TOKEN + image
        image += struct.pack('B', self._checksum256(image))

        self._write_finished_cb = write_finished_cb

        self.mem_handler.write(self, 0x00,
                               struct.unpack('B' * len(image), image))

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            self.valid = False
            logger.debug('Updating content of memory {}'.format(self.id))
            # Start reading the header
            self.mem_handler.read(self, 0, 16)

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
import struct

from .memory_element import MemoryElement

logger = logging.getLogger(__name__)


class AnchorData:
    """Holds data for one anchor"""

    def __init__(self, position=(0.0, 0.0, 0.0), is_valid=False):
        self.position = position
        self.is_valid = is_valid

    def set_from_mem_data(self, data):
        x, y, z, self.is_valid = struct.unpack('<fff?', data)
        self.position = (x, y, z)


class LocoMemory(MemoryElement):
    """Memory interface for accessing data from the Loco Positioning system"""

    SIZE_OF_FLOAT = 4
    MEM_LOCO_INFO = 0x0000
    MEM_LOCO_INFO_LEN = 1
    MEM_LOCO_ANCHOR_BASE = 0x1000
    MEM_LOCO_ANCHOR_PAGE_SIZE = 0x0100
    MEM_LOCO_PAGE_LEN = (3 * SIZE_OF_FLOAT) + 1

    def __init__(self, id, type, size, mem_handler):
        super(LocoMemory, self).__init__(id=id, type=type, size=size,
                                         mem_handler=mem_handler)
        self._update_finished_cb = None

        self.anchor_data = []
        self.nr_of_anchors = 0
        self.valid = False

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        done = False
        if mem.id == self.id:
            if addr == LocoMemory.MEM_LOCO_INFO:
                self.nr_of_anchors = data[0]
                if self.nr_of_anchors == 0:
                    done = True
                else:
                    self.anchor_data = \
                        [AnchorData() for _ in range(self.nr_of_anchors)]
                    self._request_page(0)
            else:
                page = int((addr - LocoMemory.MEM_LOCO_ANCHOR_BASE) /
                           LocoMemory.MEM_LOCO_ANCHOR_PAGE_SIZE)

                self.anchor_data[page].set_from_mem_data(data)

                next_page = page + 1
                if next_page < self.nr_of_anchors:
                    self._request_page(next_page)
                else:
                    done = True

        if done:
            self.valid = True
            if self._update_finished_cb:
                self._update_finished_cb(self)
                self._update_finished_cb = None

    def update(self, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            self.anchor_data = []
            self.nr_of_anchors = 0
            self.valid = False
            logger.debug('Updating content of memory {}'.format(self.id))

            # Start reading the header
            self.mem_handler.read(self, LocoMemory.MEM_LOCO_INFO,
                                  LocoMemory.MEM_LOCO_INFO_LEN)

    def disconnect(self):
        self._update_finished_cb = None

    def _request_page(self, page):
        addr = LocoMemory.MEM_LOCO_ANCHOR_BASE + \
            LocoMemory.MEM_LOCO_ANCHOR_PAGE_SIZE * page
        self.mem_handler.read(self, addr, LocoMemory.MEM_LOCO_PAGE_LEN)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import unittest
from test.support.asyncCallbackCaller import AsyncCallbackCaller
from unittest.mock import MagicMock

from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper
from cflib.utils.callbacks import Caller


class SyncCrazyflieTest(unittest.TestCase):

    def setUp(self):
        self.uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

        self.cf_mock = MagicMock(spec=Crazyflie)
        self.cf_mock.connected = Caller()
        self.cf_mock.connection_failed = Caller()
        self.cf_mock.disconnected = Caller()
        self.cf_mock.fully_connected = Caller()

        self.cf_mock.open_link = AsyncCallbackCaller(
            cb=self.cf_mock.connected,
            args=[self.uri],
            delay=0.2
        ).trigger

        self.close_link_mock = AsyncCallbackCaller(
            cb=self.cf_mock.disconnected,
            args=[self.uri],
            delay=0.2
        )
        self.cf_mock.close_link = self.close_link_mock.trigger

        # Register a callback to be called when connected. Use it to trigger a callback
        # to trigger the call to the param.all_updated() callback
        self.cf_mock.connected.add_callback(self._connected_callback)

        self.sut = SyncCrazyflie(self.uri, cf=self.cf_mock)

    def test_different_underlying_cf_instances(self):
        # Fixture

        # Test
        scf1 = SyncCrazyflie('uri 1')
        scf2 = SyncCrazyflie('uri 2')

        # Assert
        actual1 = scf1.cf
        actual2 = scf2.cf
        self.assertNotEqual(actual1, actual2)

    def test_open_link(self):
        # Fixture

        # Test
        self.sut.open_link()

        # Assert
        self.assertTrue(self.sut.is_link_open())

    def test_failed_open_link_raises_exception(self):
        # Fixture
        expected = 'Some error message'
        self.cf_mock.open_link = AsyncCallbackCaller(
            cb=self.cf_mock.connection_failed,
            args=[self.uri, expected]).trigger

        # Test
        try:
            self.sut.open_link()
        except Exception as e:
            actual = e.args[0]
        else:
            self.fail('Expect exception')

        # Assert
        self.assertEqual(expected, actual)
        self._assertAllCallbacksAreRemoved()

    def test_open_link_of_already_open_link_raises_exception(self):
        # Fixture
        self.sut.open_link()

        # Test
        # Assert
        with self.assertRaises(Exception):
            self.sut.open_link()

    def test_wait_for_params(self):
        # Fixture

        self.sut.open_link()

        # Test
        self.sut.wait_for_params()

        # Assert
        self.assertTrue(self.sut.is_params_updated())

    def test_do_not_wait_for_params(self):
        # Fixture

        # Test
        self.sut.open_link()

        # Assert
        self.assertFalse(self.sut.is_params_updated())

    def test_close_link(self):
        # Fixture
        self.sut.open_link()

        # Test
        self.sut.close_link()

        # Assert
        self.assertEqual(1, self.close_link_mock.call_count)
        self.assertFalse(self.sut.is_link_open())
        self._assertAllCallbacksAreRemoved()

    def test_close_link_that_is_not_open(self):
        # Fixture

        # Test
        self.sut.close_link()

        # Assert
        self.assertFalse(self.sut.is_link_open())

    def test_closed_if_connection_is_lost(self):
        # Fixture
        self.sut.open_link()

        # Test
        AsyncCallbackCaller(
            cb=self.cf_mock.disconnected,
            args=[self.uri]).call_and_wait()

        # Assert
        self.assertFalse(self.sut.is_link_open())
        self._assertAllCallbacksAreRemoved()

    def test_open_close_with_context_management(self):
        # Fixture

        # Test
        with SyncCrazyflie(self.uri, self.cf_mock) as sut:
            self.assertTrue(sut.is_link_open())

        # Assert
        self.assertEqual(1, self.close_link_mock.call_count)
        self._assertAllCallbacksAreRemoved()

    def test_wait_for_params_with_context_management(self):
        # Fixture

        # Test
        with SyncCrazyflie(self.uri, cf=self.cf_mock) as sut:
            sut.wait_for_params()
            self.assertTrue(sut.is_params_updated())

        # Assert
        self._assertAllCallbacksAreRemoved()

    def test_multiple_open_close_of_link(self):
        # Fixture

        # Test
        self.sut.open_link()
        self.sut.close_link()
        self.sut.open_link()
        self.sut.close_link()

        # Assert
        self.assertEqual(2, self.close_link_mock.call_count)
        self.assertFalse(self.sut.is_link_open())
        self._assertAllCallbacksAreRemoved()

    def test_wait_for_params_with_multiple_open_close_of_link(self):
        # Fixture

        # Test
        self.sut.open_link()
        self.assertFalse(self.sut.is_params_updated())
        self.sut.wait_for_params()
        self.assertTrue(self.sut.is_params_updated())
        self.sut.close_link()
        self.assertFalse(self.sut.is_params_updated())

        self.sut.open_link()
        self.assertFalse(self.sut.is_params_updated())
        self.sut.wait_for_params()
        self.assertTrue(self.sut.is_params_updated())
        self.sut.close_link()

        # Assert
        self.assertFalse(self.sut.is_params_updated())
        self._assertAllCallbacksAreRemoved()

    def _assertAllCallbacksAreRemoved(self):
        # Remove our probe callback
        self.cf_mock.connected.remove_callback(self._connected_callback)

        self.assertEqual(0, len(self.cf_mock.connected.callbacks))
        self.assertEqual(0, len(self.cf_mock.connection_failed.callbacks))
        self.assertEqual(0, len(self.cf_mock.disconnected.callbacks))
        self.assertEqual(0, len(self.cf_mock.fully_connected.callbacks))

    def _connected_callback(self, uri):
        AsyncCallbackCaller(
            cb=self.cf_mock.fully_connected,
            args=[self.uri],
            delay=0.2
        ).trigger()



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging

from .memory_element import MemoryElement

logger = logging.getLogger(__name__)


class LEDTimingsDriverMemory(MemoryElement):
    """Memory interface for using the LED-ring mapped memory for setting RGB
       values over time. To upload and run a show sequence of
       the LEDs in the ring"""

    def __init__(self, id, type, size, mem_handler):
        super(LEDTimingsDriverMemory, self).__init__(id=id,
                                                     type=type,
                                                     size=size,
                                                     mem_handler=mem_handler)
        self._update_finished_cb = None
        self._write_finished_cb = None

        self.timings = []

    def add(self, time, rgb, leds=0, fade=False, rotate=0):
        self.timings.append({
            'time': time,
            'rgb': rgb,
            'leds': leds,
            'fade': fade,
            'rotate': rotate
        })

    def write_data(self, write_finished_cb):
        if write_finished_cb is not None:
            self._write_finished_cb = write_finished_cb

        data = []
        for timing in self.timings:
            # In order to fit all the LEDs in one radio packet RGB565 is used
            # to compress the colors. The calculations below converts 3 bytes
            # RGB into 2 bytes RGB565. Then shifts the value of each color to
            # LSB, applies the intensity and shifts them back for correct
            # alignment on 2 bytes.
            R5 = ((int)((((int(timing['rgb']['r']) & 0xFF) * 249 + 1014) >> 11)
                        & 0x1F))
            G6 = ((int)((((int(timing['rgb']['g']) & 0xFF) * 253 + 505) >> 10)
                        & 0x3F))
            B5 = ((int)((((int(timing['rgb']['b']) & 0xFF) * 249 + 1014) >> 11)
                        & 0x1F))
            led = (int(R5) << 11) | (int(G6) << 5) | (int(B5) << 0)
            extra = ((timing['leds']) & 0x0F) | (
                (timing['fade'] << 4) & 0x10) | (
                (timing['rotate'] << 5) & 0xE0)

            if (timing['time'] & 0xFF) != 0 or led != 0 or extra != 0:
                data += [timing['time'] & 0xFF, led >> 8, led & 0xFF, extra]

        data += [0, 0, 0, 0]
        self.mem_handler.write(self, 0x00, bytearray(data), flush_queue=True)

    def write_done(self, mem, addr):
        if mem.id == self.id and self._write_finished_cb:
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
This class represents the connection to one or more Loco Positioning Anchors
"""
import struct


class LoPoAnchor():
    LPP_TYPE_POSITION = 1
    LPP_TYPE_REBOOT = 2
    LPP_TYPE_MODE = 3

    REBOOT_TO_BOOTLOADER = 0
    REBOOT_TO_FIRMWARE = 1

    MODE_TWR = 1
    MODE_TDOA = 2  # TDoA 2
    MODE_TDOA3 = 3

    def __init__(self, crazyflie):
        """
        :param crazyflie: A crazyflie object to be used as a bridge to the LoPo
         system."""
        self.crazyflie = crazyflie

    def set_position(self, anchor_id, position):
        """
        Send a packet with a position to one anchor.
        :param anchor_id: The id of the targeted anchor. This is the first byte
        of the anchor address.
        :param position: The position of the anchor, as an array
        """
        x = position[0]
        y = position[1]
        z = position[2]
        data = struct.pack('<Bfff', LoPoAnchor.LPP_TYPE_POSITION, x, y, z)

        self.crazyflie.loc.send_short_lpp_packet(anchor_id, data)

    def reboot(self, anchor_id, mode):
        data = struct.pack('<BB', LoPoAnchor.LPP_TYPE_REBOOT, mode)
        self.crazyflie.loc.send_short_lpp_packet(anchor_id, data)

    def set_mode(self, anchor_id, mode):
        """
        Send a packet to set the anchor mode. If the anchor receive the packet,
        it will change mode and resets.
        """
        data = struct.pack('<BB', LoPoAnchor.LPP_TYPE_MODE, mode)
        self.crazyflie.loc.send_short_lpp_packet(anchor_id, data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Used for sending control setpoints to the Crazyflie
"""
import logging

from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort

__author__ = 'Bitcraze AB'
__all__ = ['PlatformService']

logger = logging.getLogger(__name__)

PLATFORM_COMMAND = 0
VERSION_COMMAND = 1
APP_CHANNEL = 2

PLATFORM_SET_CONT_WAVE = 0

VERSION_GET_PROTOCOL = 0
VERSION_GET_FIRMWARE = 1

LINKSERVICE_SOURCE = 1


class PlatformService():
    """
    Used for sending control setpoints to the Crazyflie
    """

    def __init__(self, crazyflie=None):
        """
        Initialize the platform object.
        """
        self._cf = crazyflie

        self._cf.add_port_callback(CRTPPort.PLATFORM, self._platform_callback)
        self._cf.add_port_callback(CRTPPort.LINKCTRL, self._crt_service_callback)

        self._protocolVersion = -1
        self._callback = None

    def fetch_platform_informations(self, callback):
        """
        Fetch platform info from the firmware
        Should be called at the earliest in the connection sequence
        """

        self._protocolVersion = -1
        self._callback = callback

        self._request_protocol_version()

    def set_continous_wave(self, enabled):
        """
        Enable/disable the client side X-mode. When enabled this recalculates
        the setpoints before sending them to the Crazyflie.
        """
        pk = CRTPPacket()
        pk.set_header(CRTPPort.PLATFORM, PLATFORM_COMMAND)
        pk.data = (0, enabled)
        self._cf.send_packet(pk)

    def get_protocol_version(self):
        """
        Return version of the CRTP protocol
        """
        return self._protocolVersion

    def _request_protocol_version(self):
        # Sending a sink request to detect if the connected Crazyflie
        # supports protocol versioning
        pk = CRTPPacket()
        pk.set_header(CRTPPort.LINKCTRL, LINKSERVICE_SOURCE)
        pk.data = (0,)
        self._cf.send_packet(pk)
        logger.info('Request _request_protocol_version()')

    def _crt_service_callback(self, pk):
        if pk.channel == LINKSERVICE_SOURCE:
            logger.info('_crt_service_callback')
            # If the sink contains a magic string, get the protocol version,
            # otherwise -1
            if pk.data[:18].decode('utf8') == 'Bitcraze Crazyflie':
                pk = CRTPPacket()
                pk.set_header(CRTPPort.PLATFORM, VERSION_COMMAND)
                pk.data = (VERSION_GET_PROTOCOL, )
                logger.info('Request protocol version')
                self._cf.send_packet(pk)
            else:
                self._protocolVersion = -1
                logger.info('Protocol version (crt): {}'.format(self.get_protocol_version()))
                self._callback()

    def _platform_callback(self, pk):
        if pk.channel == VERSION_COMMAND:
            logger.info('_platform_callback')

            if pk.data[0] == VERSION_GET_PROTOCOL:
                self._protocolVersion = pk.data[1]
                logger.info('Protocol version (platform): {}'.format(self.get_protocol_version()))
                self._callback()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import time
from collections import namedtuple
from threading import Thread

from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie.syncLogger import SyncLogger

SwarmPosition = namedtuple('SwarmPosition', 'x y z')


class _Factory:
    """
    Default Crazyflie factory class.
    """

    def construct(self, uri):
        return SyncCrazyflie(uri)


class CachedCfFactory:
    """
    Factory class that creates Crazyflie instances with TOC caching
    to reduce connection time.
    """

    def __init__(self, ro_cache=None, rw_cache=None):
        self.ro_cache = ro_cache
        self.rw_cache = rw_cache

    def construct(self, uri):
        cf = Crazyflie(ro_cache=self.ro_cache, rw_cache=self.rw_cache)
        return SyncCrazyflie(uri, cf=cf)


class Swarm:
    """
    Runs a swarm of Crazyflies. It implements a functional-ish style of
    sequential or parallel actions on all individuals of the swarm.

    When the swarm is connected, a link is opened to each Crazyflie through
    SyncCrazyflie instances. The instances are maintained by the class and are
    passed in as the first argument in swarm wide actions.
    """

    def __init__(self, uris, factory=_Factory()):
        """
        Constructs a Swarm instance and instances used to connect to the
        Crazyflies

        :param uris: A set of uris to use when connecting to the Crazyflies in
        the swarm
        :param factory: A factory class used to create the instances that are
         used to open links to the Crazyflies. Mainly used for unit testing.
        """
        self._cfs = {}
        self._is_open = False
        self._positions = dict()

        for uri in uris:
            self._cfs[uri] = factory.construct(uri)

    def open_links(self):
        """
        Open links to all individuals in the swarm
        """
        if self._is_open:
            raise Exception('Already opened')

        try:
            self.parallel_safe(lambda scf: scf.open_link())
            self._is_open = True
        except Exception as e:
            self.close_links()
            raise e

    def close_links(self):
        """
        Close all open links
        """
        for uri, cf in self._cfs.items():
            cf.close_link()

        self._is_open = False

    def __enter__(self):
        self.open_links()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_links()

    def __get_estimated_position(self, scf):
        log_config = LogConfig(name='stateEstimate', period_in_ms=10)
        log_config.add_variable('stateEstimate.x', 'float')
        log_config.add_variable('stateEstimate.y', 'float')
        log_config.add_variable('stateEstimate.z', 'float')

        with SyncLogger(scf, log_config) as logger:
            for entry in logger:
                x = entry[1]['stateEstimate.x']
                y = entry[1]['stateEstimate.y']
                z = entry[1]['stateEstimate.z']
                self._positions[scf.cf.link_uri] = SwarmPosition(x, y, z)
                break

    def get_estimated_positions(self):
        """
        Return a `dict`, keyed by URI and with the SwarmPosition namedtuples as
        value, with the estimated (x, y, z) of each Crazyflie in the swarm.
        """
        self.parallel_safe(self.__get_estimated_position)
        return self._positions

    def __wait_for_position_estimator(self, scf):
        log_config = LogConfig(name='Kalman Variance', period_in_ms=500)
        log_config.add_variable('kalman.varPX', 'float')
        log_config.add_variable('kalman.varPY', 'float')
        log_config.add_variable('kalman.varPZ', 'float')

        var_y_history = [1000] * 10
        var_x_history = [1000] * 10
        var_z_history = [1000] * 10

        threshold = 0.001

        with SyncLogger(scf, log_config) as logger:
            for log_entry in logger:
                data = log_entry[1]

                var_x_history.append(data['kalman.varPX'])
                var_x_history.pop(0)
                var_y_history.append(data['kalman.varPY'])
                var_y_history.pop(0)
                var_z_history.append(data['kalman.varPZ'])
                var_z_history.pop(0)

                min_x = min(var_x_history)
                max_x = max(var_x_history)
                min_y = min(var_y_history)
                max_y = max(var_y_history)
                min_z = min(var_z_history)
                max_z = max(var_z_history)

                if (max_x - min_x) < threshold and (
                        max_y - min_y) < threshold and (
                        max_z - min_z) < threshold:
                    break

    def __reset_estimator(self, scf):
        cf = scf.cf
        cf.param.set_value('kalman.resetEstimation', '1')
        time.sleep(0.1)
        cf.param.set_value('kalman.resetEstimation', '0')
        self.__wait_for_position_estimator(scf)

    def reset_estimators(self):
        """
        Reset estimator on all members of the swarm and wait for a stable
        positions. Blocks until position estimators finds a position.
        """
        self.parallel_safe(self.__reset_estimator)

    def sequential(self, func, args_dict=None):
        """
        Execute a function for all Crazyflies in the swarm, in sequence.

        The first argument of the function that is passed in will be a
        SyncCrazyflie instance connected to the Crazyflie to operate on.
        A list of optional parameters (per Crazyflie) may follow defined by
        the `args_dict`. The dictionary is keyed on URI and has a list of
        parameters as value.

        Example:
        ```python
        def my_function(scf, optional_param0, optional_param1)
            ...

        args_dict = {
            URI0: [optional_param0_cf0, optional_param1_cf0],
            URI1: [optional_param0_cf1, optional_param1_cf1],
            ...
        }


        swarm.sequential(my_function, args_dict)
        ```

        :param func: The function to execute
        :param args_dict: Parameters to pass to the function
        """
        for uri, cf in self._cfs.items():
            args = self._process_args_dict(cf, uri, args_dict)
            func(*args)

    def parallel(self, func, args_dict=None):
        """
        Execute a function for all Crazyflies in the swarm, in parallel.
        One thread per Crazyflie is started to execute the function. The
        threads are joined at the end. Exceptions raised by the threads are
        ignored.

        For a more detailed description of the arguments, see `sequential()`

        :param func: The function to execute
        :param args_dict: Parameters to pass to the function
        """
        try:
            self.parallel_safe(func, args_dict)
        except Exception:
            pass

    def parallel_safe(self, func, args_dict=None):
        """
        Execute a function for all Crazyflies in the swarm, in parallel.
        One thread per Crazyflie is started to execute the function. The
        threads are joined at the end and if one or more of the threads raised
        an exception this function will also raise an exception.

        For a more detailed description of the arguments, see `sequential()`

        :param func: The function to execute
        :param args_dict: Parameters to pass to the function
        """
        threads = []
        reporter = self.Reporter()

        for uri, scf in self._cfs.items():
            args = [func, reporter] + \
                self._process_args_dict(scf, uri, args_dict)

            thread = Thread(target=self._thread_function_wrapper, args=args)
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        if reporter.is_error_reported():
            first_error = reporter.errors[0]
            raise Exception('One or more threads raised an exception when '
                            'executing parallel task') from first_error

    def _thread_function_wrapper(self, *args):
        reporter = None
        try:
            func = args[0]
            reporter = args[1]
            func(*args[2:])
        except Exception as e:
            if reporter:
                reporter.report_error(e)

    def _process_args_dict(self, scf, uri, args_dict):
        args = [scf]

        if args_dict:
            args += args_dict[uri]

        return args

    class Reporter:
        def __init__(self):
            self.error_reported = False
            self._errors = []

        @property
        def errors(self):
            return self._errors

        def report_error(self, e):
            self.error_reported = True
            self._errors.append(e)

        def is_error_reported(self):
            return self.error_reported



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
The MotionCommander is used to make it easy to write scripts that moves the
Crazyflie around. Some sort of positioning support is required, for instance
the Flow deck.

The motion commander uses velocity setpoints and does not have a notion of
absolute position, the error in position will accumulate over time.

The API contains a set of primitives that are easy to understand and use, such
as "go forward" or "turn around".

There are two flavors of primitives, one that is blocking and returns when
a motion is completed, while the other starts a motion and returns immediately.
In the second variation the user has to stop or change the motion when
appropriate by issuing new commands.

The MotionCommander can be used as context manager using the with keyword. In
this mode of operation takeoff and landing is executed when the context is
created/closed.
"""
import math
import time
from queue import Empty
from queue import Queue
from threading import Thread

from cflib.crazyflie.syncCrazyflie import SyncCrazyflie


class MotionCommander:
    """The motion commander"""
    VELOCITY = 0.2
    RATE = 360.0 / 5

    def __init__(self, crazyflie, default_height=0.3):
        """
        Construct an instance of a MotionCommander

        :param crazyflie: A Crazyflie or SyncCrazyflie instance
        :param default_height: The default height to fly at
        """
        if isinstance(crazyflie, SyncCrazyflie):
            self._cf = crazyflie.cf
        else:
            self._cf = crazyflie

        self.default_height = default_height

        self._is_flying = False
        self._thread = None

    # Distance based primitives

    def take_off(self, height=None, velocity=VELOCITY):
        """
        Takes off, that is starts the motors, goes straight up and hovers.
        Do not call this function if you use the with keyword. Take off is
        done automatically when the context is created.

        :param height: The height (meters) to hover at. None uses the default
                       height set when constructed.
        :param velocity: The velocity (meters/second) when taking off
        :return:
        """
        if self._is_flying:
            raise Exception('Already flying')

        if not self._cf.is_connected():
            raise Exception('Crazyflie is not connected')

        self._is_flying = True
        self._reset_position_estimator()

        self._thread = _SetPointThread(self._cf)
        self._thread.start()

        if height is None:
            height = self.default_height

        self.up(height, velocity)

    def land(self, velocity=VELOCITY):
        """
        Go straight down and turn off the motors.

        Do not call this function if you use the with keyword. Landing is
        done automatically when the context goes out of scope.

        :param velocity: The velocity (meters/second) when going down
        :return:
        """
        if self._is_flying:
            self.down(self._thread.get_height(), velocity)

            self._thread.stop()
            self._thread = None

            self._cf.commander.send_stop_setpoint()
            self._is_flying = False

    def __enter__(self):
        self.take_off()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.land()

    def left(self, distance_m, velocity=VELOCITY):
        """
        Go left

        :param distance_m: The distance to travel (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.move_distance(0.0, distance_m, 0.0, velocity)

    def right(self, distance_m, velocity=VELOCITY):
        """
        Go right

        :param distance_m: The distance to travel (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.move_distance(0.0, -distance_m, 0.0, velocity)

    def forward(self, distance_m, velocity=VELOCITY):
        """
        Go forward

        :param distance_m: The distance to travel (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.move_distance(distance_m, 0.0, 0.0, velocity)

    def back(self, distance_m, velocity=VELOCITY):
        """
        Go backwards

        :param distance_m: The distance to travel (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.move_distance(-distance_m, 0.0, 0.0, velocity)

    def up(self, distance_m, velocity=VELOCITY):
        """
        Go up

        :param distance_m: The distance to travel (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.move_distance(0.0, 0.0, distance_m, velocity)

    def down(self, distance_m, velocity=VELOCITY):
        """
        Go down

        :param distance_m: The distance to travel (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.move_distance(0.0, 0.0, -distance_m, velocity)

    def turn_left(self, angle_degrees, rate=RATE):
        """
        Turn to the left, staying on the spot

        :param angle_degrees: How far to turn (degrees)
        :param rate: The turning speed (degrees/second)
        :return:
        """
        flight_time = angle_degrees / rate

        self.start_turn_left(rate)
        time.sleep(flight_time)
        self.stop()

    def turn_right(self, angle_degrees, rate=RATE):
        """
        Turn to the right, staying on the spot

        :param angle_degrees: How far to turn (degrees)
        :param rate: The turning speed (degrees/second)
        :return:
        """
        flight_time = angle_degrees / rate

        self.start_turn_right(rate)
        time.sleep(flight_time)
        self.stop()

    def circle_left(self, radius_m, velocity=VELOCITY, angle_degrees=360.0):
        """
        Go in circle, counter clock wise

        :param radius_m: The radius of the circle (meters)
        :param velocity: The velocity along the circle (meters/second)
        :param angle_degrees: How far to go in the circle (degrees)
        :return:
        """
        distance = 2 * radius_m * math.pi * angle_degrees / 360.0
        flight_time = distance / velocity

        self.start_circle_left(radius_m, velocity)
        time.sleep(flight_time)
        self.stop()

    def circle_right(self, radius_m, velocity=VELOCITY, angle_degrees=360.0):
        """
        Go in circle, clock wise

        :param radius_m: The radius of the circle (meters)
        :param velocity: The velocity along the circle (meters/second)
        :param angle_degrees: How far to go in the circle (degrees)
        :return:
        """
        distance = 2 * radius_m * math.pi * angle_degrees / 360.0
        flight_time = distance / velocity

        self.start_circle_right(radius_m, velocity)
        time.sleep(flight_time)
        self.stop()

    def move_distance(self, distance_x_m, distance_y_m, distance_z_m,
                      velocity=VELOCITY):
        """
        Move in a straight line.
        positive X is forward
        positive Y is left
        positive Z is up

        :param distance_x_m: The distance to travel along the X-axis (meters)
        :param distance_y_m: The distance to travel along the Y-axis (meters)
        :param distance_z_m: The distance to travel along the Z-axis (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        distance = math.sqrt(distance_x_m * distance_x_m +
                             distance_y_m * distance_y_m +
                             distance_z_m * distance_z_m)
        flight_time = distance / velocity

        velocity_x = velocity * distance_x_m / distance
        velocity_y = velocity * distance_y_m / distance
        velocity_z = velocity * distance_z_m / distance

        self.start_linear_motion(velocity_x, velocity_y, velocity_z)
        time.sleep(flight_time)
        self.stop()

    # Velocity based primitives

    def start_left(self, velocity=VELOCITY):
        """
        Start moving left. This function returns immediately.

        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.start_linear_motion(0.0, velocity, 0.0)

    def start_right(self, velocity=VELOCITY):
        """
        Start moving right. This function returns immediately.

        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.start_linear_motion(0.0, -velocity, 0.0)

    def start_forward(self, velocity=VELOCITY):
        """
        Start moving forward. This function returns immediately.

        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.start_linear_motion(velocity, 0.0, 0.0)

    def start_back(self, velocity=VELOCITY):
        """
        Start moving backwards. This function returns immediately.

        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.start_linear_motion(-velocity, 0.0, 0.0)

    def start_up(self, velocity=VELOCITY):
        """
        Start moving up. This function returns immediately.

        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.start_linear_motion(0.0, 0.0, velocity)

    def start_down(self, velocity=VELOCITY):
        """
        Start moving down. This function returns immediately.

        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        self.start_linear_motion(0.0, 0.0, -velocity)

    def stop(self):
        """
        Stop any motion and hover.

        :return:
        """
        self._set_vel_setpoint(0.0, 0.0, 0.0, 0.0)

    def start_turn_left(self, rate=RATE):
        """
        Start turning left. This function returns immediately.

        :param rate: The angular rate (degrees/second)
        :return:
        """
        self._set_vel_setpoint(0.0, 0.0, 0.0, -rate)

    def start_turn_right(self, rate=RATE):
        """
        Start turning right. This function returns immediately.

        :param rate: The angular rate (degrees/second)
        :return:
        """
        self._set_vel_setpoint(0.0, 0.0, 0.0, rate)

    def start_circle_left(self, radius_m, velocity=VELOCITY):
        """
        Start a circular motion to the left. This function returns immediately.

        :param radius_m: The radius of the circle (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        circumference = 2 * radius_m * math.pi
        rate = 360.0 * velocity / circumference

        self._set_vel_setpoint(velocity, 0.0, 0.0, -rate)

    def start_circle_right(self, radius_m, velocity=VELOCITY):
        """
        Start a circular motion to the right. This function returns immediately

        :param radius_m: The radius of the circle (meters)
        :param velocity: The velocity of the motion (meters/second)
        :return:
        """
        circumference = 2 * radius_m * math.pi
        rate = 360.0 * velocity / circumference

        self._set_vel_setpoint(velocity, 0.0, 0.0, rate)

    def start_linear_motion(self, velocity_x_m, velocity_y_m, velocity_z_m, rate_yaw=0.0):
        """
        Start a linear motion with an optional yaw rate input. This function returns immediately.

        positive X is forward
        positive Y is left
        positive Z is up

        :param velocity_x_m: The velocity along the X-axis (meters/second)
        :param velocity_y_m: The velocity along the Y-axis (meters/second)
        :param velocity_z_m: The velocity along the Z-axis (meters/second)
        :param rate: The angular rate (degrees/second)
        :return:
        """
        self._set_vel_setpoint(
            velocity_x_m, velocity_y_m, velocity_z_m, rate_yaw)

    def _set_vel_setpoint(self, velocity_x, velocity_y, velocity_z, rate_yaw):
        if not self._is_flying:
            raise Exception('Can not move on the ground. Take off first!')
        self._thread.set_vel_setpoint(
            velocity_x, velocity_y, velocity_z, rate_yaw)

    def _reset_position_estimator(self):
        self._cf.param.set_value('kalman.resetEstimation', '1')
        time.sleep(0.1)
        self._cf.param.set_value('kalman.resetEstimation', '0')
        time.sleep(2)


class _SetPointThread(Thread):
    TERMINATE_EVENT = 'terminate'
    UPDATE_PERIOD = 0.2
    ABS_Z_INDEX = 3

    def __init__(self, cf, update_period=UPDATE_PERIOD):
        Thread.__init__(self)
        self.update_period = update_period

        self._queue = Queue()
        self._cf = cf

        self._hover_setpoint = [0.0, 0.0, 0.0, 0.0]

        self._z_base = 0.0
        self._z_velocity = 0.0
        self._z_base_time = 0.0

    def stop(self):
        """
        Stop the thread and wait for it to terminate

        :return:
        """
        self._queue.put(self.TERMINATE_EVENT)
        self.join()

    def set_vel_setpoint(self, velocity_x, velocity_y, velocity_z, rate_yaw):
        """Set the velocity setpoint to use for the future motion"""
        self._queue.put((velocity_x, velocity_y, velocity_z, rate_yaw))

    def get_height(self):
        """
        Get the current height of the Crazyflie.

        :return: The height (meters)
        """
        return self._hover_setpoint[self.ABS_Z_INDEX]

    def run(self):
        while True:
            try:
                event = self._queue.get(block=True, timeout=self.update_period)
                if event == self.TERMINATE_EVENT:
                    return

                self._new_setpoint(*event)
            except Empty:
                pass

            self._update_z_in_setpoint()
            self._cf.commander.send_hover_setpoint(*self._hover_setpoint)

    def _new_setpoint(self, velocity_x, velocity_y, velocity_z, rate_yaw):
        self._z_base = self._current_z()
        self._z_velocity = velocity_z
        self._z_base_time = time.time()

        self._hover_setpoint = [velocity_x, velocity_y, rate_yaw, self._z_base]

    def _update_z_in_setpoint(self):
        self._hover_setpoint[self.ABS_Z_INDEX] = self._current_z()

    def _current_z(self):
        now = time.time()
        return self._z_base + self._z_velocity * (now - self._z_base_time)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import unittest

from cflib.localization import LighthouseBsGeoEstimator


class TestLighthouseBsGeoEstimator(unittest.TestCase):
    def setUp(self):

        self.sut = LighthouseBsGeoEstimator()

    def test_that_sanity_check_finds_coordinate_out_of_bounds(self):
        # Fixture
        pos_bs_in_cf_coord = [0, -20, 0]

        # Test
        actual = self.sut.sanity_check_result(pos_bs_in_cf_coord)

        # Assert
        self.assertFalse(actual)

    def test_that_sanity_check_passes_ok_position(self):
        # Fixture
        pos_bs_in_cf_coord = [0, 1, 2]

        # Test
        actual = self.sut.sanity_check_result(pos_bs_in_cf_coord)

        # Assert
        self.assertTrue(actual)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016-2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
This class provides synchronous access to log data from the Crazyflie.

It acts as an iterator and returns the next value on each iteration.
If no value is available it blocks until log data is received again.
"""
from queue import Queue

from cflib.crazyflie.syncCrazyflie import SyncCrazyflie


class SyncLogger:
    DISCONNECT_EVENT = 'DISCONNECT_EVENT'

    def __init__(self, crazyflie, log_config):
        """
        Construct an instance of a SyncLogger

        Takes an Crazyflie or SyncCrazyflie instance and one log configuration
        or an array of log configurations
        """
        if isinstance(crazyflie, SyncCrazyflie):
            self._cf = crazyflie.cf
        else:
            self._cf = crazyflie

        if isinstance(log_config, list):
            self._log_config = log_config
        else:
            self._log_config = [log_config]

        self._queue = Queue()

        self._is_connected = False

    def connect(self):
        if self._is_connected:
            raise Exception('Already connected')

        self._cf.disconnected.add_callback(self._disconnected)
        for config in self._log_config:
            self._cf.log.add_config(config)
            config.data_received_cb.add_callback(self._log_callback)
            config.start()

        self._is_connected = True

    def disconnect(self):
        if self._is_connected:
            for config in self._log_config:
                config.stop()
                config.delete()

                config.data_received_cb.remove_callback(
                    self._log_callback)

            self._cf.disconnected.remove_callback(self._disconnected)

            self._is_connected = False

    def is_connected(self):
        return self._is_connected

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if not self._is_connected:
            raise StopIteration

        data = self._queue.get()

        if data == self.DISCONNECT_EVENT:
            self._queue.empty()
            raise StopIteration

        return data

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()
        self._queue.empty()

    def _log_callback(self, ts, data, logblock):
        self._queue.put((ts, data, logblock))

    def _disconnected(self, link_uri):
        self.disconnect()
        self._queue.put(self.DISCONNECT_EVENT)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.optimize

from cflib.localization.lighthouse_types import LhBsCfPoses
from cflib.localization.lighthouse_types import LhCfPoseSample
from cflib.localization.lighthouse_types import Pose


class LighthouseGeometrySolution:
    """
    Represents a solution from the geometry solver.

    Some data in the object is also used during the solving process for context.
    """

    def __init__(self) -> None:
        # Nr of parameters in a rotation vector
        self.len_rot_vec = 3
        # Nr of parameters
        self.len_pose = 6

        # Nr of base stations
        self.n_bss: int = None
        # Nr of parametrs per base station
        self.n_params_per_bs = self.len_pose

        # Nr of sampled Crazyflie poses
        self.n_cfs: int = None
        # Nr of sampled Crazyflie poses used in the parameter set
        self.n_cfs_in_params: int = None
        # Nr of parameters per Crazyflie pose
        self.n_params_per_cf = self.len_pose

        # Nr of sensors
        self.n_sensors: int = None

        # The maximum nr of iterations to execute when solving the system
        self.max_nr_iter = 100

        self.bs_id_to_index: dict[int, int] = {}
        self.bs_index_to_id: dict[int, int] = {}

        # The solution ######################

        # The estimated poses of the base stations
        self.bs_poses: dict[int, Pose] = {}

        # The estimated poses of the CF samples
        self.cf_poses: list[Pose] = []

        # Estimated error for each base station in each sample
        self.estimated_errors: list[dict[int, float]] = []

        # Information about errors in the solution
        self.error_info = {}

        # Indicates if the solution coverged (True).
        # If it did not converge, the solution is probably not good enough to use
        self.success = False


class LighthouseGeometrySolver:
    """
    Finds the poses of base stations and Crazyflie samples given a list of matched samples.
    The solver is iterative and uses least squares fitting to minimize the distance from
    the lighthouse sensors to each "ray" measured in the samples.

    The equation system that is solved is defined as:

    Columns are the estimated poses (what we solve for). Each pose is composed of 6 numbers (often referred to as
    parameters in the code): rotation vector (3) and position (3).

    Rows are representing one angle from one base station. The number of rows for each sample is given by the
    number of bs in the sample * n_sensors * 2.

    An examples matrix:

                        bs0_pose, bs1_pose, bs2_pose, bs3_pose, cf1_pose, cf2_pose, ...

    cf0/bs2/sens0/ang0                        X
    cf0/bs2/sens0/ang1                        X
    cf0/bs2/sens1/ang0                        X
    cf0/bs2/sens1/ang1                        X
    ...
    cf0/bs3/sens0/ang0                                  X
    cf0/bs3/sens0/ang1                                  X
    cf0/bs3/sens1/ang0                                  X
    cf0/bs3/sens1/ang1                                  X
    ...
    cf1/bs1/sens0/ang0              X                             X
    cf1/bs1/sens0/ang1              X                             X
    cf1/bs1/sens1/ang0              X                             X
    cf1/bs1/sens1/ang1              X                             X
    ...
    cf1/bs2/sens0/ang0                        X                   X
    cf1/bs2/sens0/ang1                        X                   X
    cf1/bs2/sens1/ang0                        X                   X
    cf1/bs2/sens1/ang1                        X                   X
    ...
    cf2/bs1/sens0/ang0              X                                      X
    cf2/bs1/sens0/ang1              X                                      X
    cf2/bs1/sens1/ang0              X                                      X
    cf2/bs1/sens1/ang1              X                                      X
    ...
    cf2/bs3/sens0/ang0                                  X                  X
    cf2/bs3/sens0/ang1                                  X                  X
    cf2/bs3/sens1/ang0                                  X                  X
    cf2/bs3/sens1/ang1                                  X                  X
    ...
    """

    @classmethod
    def solve(cls, initial_guess: LhBsCfPoses, matched_samples: list[LhCfPoseSample],
              sensor_positions: npt.ArrayLike) -> LighthouseGeometrySolution:
        """
        Solve for the pose of base stations and CF samples.
        The pose of the CF in sample 0 defines the global reference frame.

        Iteration is terminated acceptable solution is found. If no solution is found after a fixed number of iterations
        the solver is terminated. The success member of the result will indicate if a solution was found or not.

        :param initial_guess: Initial guess for the base stations and CF sample poses
        :param matched_samples: List of matched samples.
        :param sensor_positions: Sensor positions (3D), in the CF reference frame
        :return: an instance of LighthouseGeometrySolution
        """
        solution = LighthouseGeometrySolution()

        solution.n_bss = len(initial_guess.bs_poses)
        solution.n_cfs = len(matched_samples)
        solution.n_cfs_in_params = len(matched_samples) - 1
        solution.n_sensors = len(sensor_positions)
        solution.bs_id_to_index, solution.bs_index_to_id = cls._create_bs_map(initial_guess.bs_poses)

        target_angles = cls._populate_target_angles(matched_samples)
        idx_agl_pr_to_bs, idx_agl_pr_to_cf, idx_agl_pr_to_sens_pos, jac_sparsity = cls._populate_indexes_and_jacobian(
            matched_samples, solution)
        params_bs, params_cfs = cls._populate_initial_guess(initial_guess, solution)

        # Extra arguments passed on to calc_residual()
        args = (solution, idx_agl_pr_to_bs, idx_agl_pr_to_cf, idx_agl_pr_to_sens_pos, target_angles, sensor_positions)

        # Vector to optimize. Composed of base station parameters followed by cf parameters
        x0 = np.hstack((params_bs.ravel(), params_cfs.ravel()))

        result = scipy.optimize.least_squares(cls._calc_residual,
                                              x0,
                                              verbose=0,
                                              jac_sparsity=jac_sparsity,
                                              x_scale='jac',
                                              ftol=1e-8,
                                              method='trf',
                                              max_nfev=solution.max_nr_iter,
                                              args=args)

        cls._condense_results(result, solution, matched_samples)
        return solution

    @classmethod
    def _populate_target_angles(cls, matched_samples: list[LhCfPoseSample]) -> npt.NDArray:
        """
        A np.array of all measured angles, the target angles
        """
        result = []
        for sample in matched_samples:
            for bs_id, angles in sorted(sample.angles_calibrated.items()):
                result += angles.angle_list().tolist()

        return np.array(result)

    @classmethod
    def _populate_indexes_and_jacobian(cls, matched_samples: list[LhCfPoseSample], defs: LighthouseGeometrySolution
                                       ) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
        """
        To speed up calculations all operations in the iteration phase are done on np.arrays of equal length (ish),
        the numpy flavour of parallell work. Some data is reused in multiple equations (for instance sensor
        positions) and to avoid copying of data we use np.arrays as indexes. This method creates the necessary
        indexes.

        Since the equation system we are solving is sparse, we can also do a second optimization and only calculate
        the parts of the matrix that are non-zero. We have to provide a matrix containing 1 in the positions where
        the jacobian is non-zero, this matrix is also generated here.
        """
        index_angle_pair_to_bs = []
        index_angle_pair_to_cf = []
        index_angle_pair_to_sensor_base = []

        # Note: Indexes are for angle pairs, that is one set of indexes for two equations in the matrix.
        # Each set of indexes will result in two angles, horizontal and vertical, which means there is one set of
        # indexes per sensor.

        for cf_i, sample in enumerate(matched_samples):
            for bs_id in sorted(sample.angles_calibrated.keys()):
                bs_index = defs.bs_id_to_index[bs_id]
                for sensor_i in range(defs.n_sensors):
                    index_angle_pair_to_cf.append(cf_i)
                    index_angle_pair_to_bs.append(bs_index)
                    index_angle_pair_to_sensor_base.append(sensor_i)

        # Length of residual vector
        len_residual_vec = len(index_angle_pair_to_cf) * 2

        # Length of param vector
        len_param_vec = defs.n_bss * defs.n_params_per_bs + defs.n_cfs_in_params * defs.n_params_per_cf

        # The jac_sparsity matrix should have ones in all locations where data is used in the equations
        jac_sparsity = scipy.sparse.lil_matrix((len_residual_vec, len_param_vec), dtype=int)
        row_i = 0
        n_tot_bs_params = defs.n_bss * defs.n_params_per_bs
        for cf_i, sample in enumerate(matched_samples):
            for bs_id in sorted(sample.angles_calibrated.keys()):
                bs_index = defs.bs_id_to_index[bs_id]
                for sensor_i in range(defs.n_sensors * 2):
                    # Add bs parameters
                    first = bs_index * defs.n_params_per_bs
                    for i in range(first, first + defs.n_params_per_bs):
                        jac_sparsity[row_i, i] = 1
                    # Add cf parameters
                    if cf_i > 0:
                        first = n_tot_bs_params + (cf_i - 1) * defs.n_params_per_cf
                        for i in range(first, first + defs.n_params_per_cf):
                            jac_sparsity[row_i, i] = 1

                    row_i += 1

        return (np.array(index_angle_pair_to_bs),
                np.array(index_angle_pair_to_cf),
                np.array(index_angle_pair_to_sensor_base),
                jac_sparsity)

    @classmethod
    def _populate_initial_guess(cls, initial_guess: LhBsCfPoses,
                                defs: LighthouseGeometrySolution) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Generate parameters for base stations and CFs, this is the initial guess we start to iterate from.
        """
        params_bs = np.zeros((defs.n_bss, defs.n_params_per_bs))
        for bs_id, pose in initial_guess.bs_poses.items():
            params_bs[defs.bs_id_to_index[bs_id], :] = cls._pose_to_params(pose)

        # Skip the first CF pose, it is the definition of the origin and is not a parameter
        params_cfs = np.zeros((defs.n_cfs_in_params, defs.n_params_per_cf))
        for index, inital_est_pose in enumerate(initial_guess.cf_poses[1:]):
            params_cfs[index, :] = cls._pose_to_params(inital_est_pose)

        return params_bs, params_cfs

    @classmethod
    def _params_to_struct(cls, params, defs: LighthouseGeometrySolution):
        """
        Convert the params list to two arrays, one for base stations and one for CFs
        """
        bs_param_count = defs.n_bss * defs.n_params_per_bs
        params_bs_poses = params[:bs_param_count].reshape((defs.n_bss, defs.n_params_per_bs))

        params_cf_poses = params[bs_param_count:].reshape((defs.n_cfs_in_params, defs.n_params_per_cf))

        return params_bs_poses, params_cf_poses

    @classmethod
    def _calc_residual(cls, params, defs: LighthouseGeometrySolution, index_angle_pair_to_bs, index_angle_pair_to_cf,
                       index_angle_pair_to_sensor_base, target_angles, sensor_positions):
        """
        Calculate the residual for a set of parameters. The residual is defined as the distance from a sensor to the
        plane given by a measured base station angle.

        :param params: list of parameters for base stations and CFs
        :param defs: information about the context
        :param index_angle_pair_to_bs: index array to index into the base station part of the parameter set
        :param index_angle_pair_to_cf: index array to index into the CF part of the parameter set
        :param index_angle_pair_to_sensor_base: index array to index into the sensor position array
        :param target_angles: the measured angles
        :param sensor_positions: Array with sensor positions
        :return: Array with residuals
        """
        bss, cfs = cls._params_to_struct(params, defs)

        # The first CF pose is defining the origin and is added here
        cfs_full = np.concatenate((np.zeros((1, defs.n_params_per_cf), dtype=float), cfs))

        angle_pairs = cls._poses_to_angle_pairs(bss, cfs_full, sensor_positions, index_angle_pair_to_bs,
                                                index_angle_pair_to_cf, index_angle_pair_to_sensor_base, defs)
        angles = np.ravel(angle_pairs)

        diff = angles - target_angles

        # Calculate the error at the CF positions
        distances_to_cfs = np.repeat(np.linalg.norm(
            bss[index_angle_pair_to_bs][:, 3:] - cfs_full[index_angle_pair_to_cf][:, 3:], axis=1), 2)
        residual = np.tan(diff) * distances_to_cfs

        return residual

    @classmethod
    def _poses_to_angle_pairs(cls, bss, cf_poses, sensor_base_pos, index_angle_pair_to_bs, index_angle_pair_to_cf,
                              index_angle_pair_to_sensor_base, defs: LighthouseGeometrySolution):
        pairs = cls._calc_angle_pairs(bss[index_angle_pair_to_bs], cf_poses[index_angle_pair_to_cf],
                                      sensor_base_pos[index_angle_pair_to_sensor_base], defs)
        return pairs

    @classmethod
    def _calc_angle_pairs(cls, bs_p_a, cf_p_a, sens_pos_p_a, defs: LighthouseGeometrySolution):
        """
        Calculate angle pairs based on base station poses, cf poses and sensor positions

        :param bs_p_a: Poses base stations
        :param cf_p_a: Poses CFs
        :param sens_pos_p_a: Sensor positions
        :return: angle pairs

        All lists are equally long, one entry per output angle pair
        """
        sensor_points = cls._rotate_translate(sens_pos_p_a, cf_p_a[:, :defs.len_rot_vec], cf_p_a[:, defs.len_rot_vec:])

        # translate and inverse rotate (-rotation vector == inverse rotation)
        points_bs_ref = cls._rotate_translate(sensor_points - bs_p_a[:, defs.len_rot_vec:defs.n_params_per_bs],
                                              -bs_p_a[:, :defs.len_rot_vec],
                                              np.zeros_like(bs_p_a[:, defs.len_rot_vec:defs.n_params_per_bs]))

        angle_pair = np.arctan2(points_bs_ref[:, 1:3], points_bs_ref[:, 0, np.newaxis])
        return angle_pair

    @classmethod
    def _rotate_translate(cls, points, rot_vecs, translations):
        """Rotate points by given rotation vectors and translate

        Rodrigues' rotation formula is used.
        """
        theta = np.linalg.norm(rot_vecs, axis=1)[:, np.newaxis]
        with np.errstate(invalid='ignore'):
            v = rot_vecs / theta
            v = np.nan_to_num(v)
        dot = np.sum(points * v, axis=1)[:, np.newaxis]
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)

        return cos_theta * points + sin_theta * np.cross(v, points) + dot * (1 - cos_theta) * v + translations

    @classmethod
    def _pose_to_params(cls, pose: Pose) -> npt.NDArray:
        """
        Convert from Pose to the array format used in the solver
        """
        return np.concatenate((pose.rot_vec, pose.translation))

    @classmethod
    def _params_to_pose(cls, params: npt.ArrayLike, defs: LighthouseGeometrySolution) -> Pose:
        """
        Convert from the array format used in the solver to Pose
        """
        r_vec = params[:defs.len_rot_vec]
        t = params[defs.len_rot_vec:defs.len_pose]
        return Pose.from_rot_vec(R_vec=r_vec, t_vec=t)

    @classmethod
    def _create_bs_map(cls, initial_guess_bs_poses: dict[int, Pose]) -> tuple[dict[int, int], dict[int, int]]:
        """
        We might have gaps in the list of base station ids that is used in the system, use an index instead
        when refering to a base station. This method creates dictionaries to go from index to base station id,
        or the other way around.

        Base station ids are indexed in an increasing order which means that sorting keys will result
        in sorted indexes as well.
        """
        bs_id_to_index = {}
        bs_index_to_id = {}

        for index, id in enumerate(sorted(initial_guess_bs_poses.keys())):
            bs_id_to_index[id] = index
            bs_index_to_id[index] = id

        return bs_id_to_index, bs_index_to_id

    @classmethod
    def _condense_results(cls, lsq_result, solution: LighthouseGeometrySolution,
                          matched_samples: list[LhCfPoseSample]) -> None:
        bss, cf_poses = cls._params_to_struct(lsq_result.x, solution)

        # Extract CF pose estimates
        # First pose (origin) is not in the parameter list
        solution.cf_poses.append(Pose())
        for i in range(len(matched_samples) - 1):
            solution.cf_poses.append(cls._params_to_pose(cf_poses[i], solution))

        # Extract base station pose estimates
        for index, pose in enumerate(bss):
            bs_id = solution.bs_index_to_id[index]
            solution.bs_poses[bs_id] = cls._params_to_pose(pose, solution)

        solution.success = lsq_result.success

        # Extract the error for each CF pose
        residuals = lsq_result.fun
        i = 0
        for sample in matched_samples:
            sample_errors = {}
            for bs_id in sorted(sample.angles_calibrated.keys()):
                sample_errors[bs_id] = np.linalg.norm(residuals[i:i + 2])
                i += solution.n_sensors * 2
            solution.estimated_errors.append(sample_errors)

        solution.error_info = cls._aggregate_error_info(solution.estimated_errors)

    @classmethod
    def _aggregate_error_info(cls, estimated_errors: list[dict[int, float]]) -> dict[str, float]:
        error_per_bs = {}
        errors = []
        for sample_errors in estimated_errors:
            for bs_id, error in sample_errors.items():
                if bs_id not in error_per_bs:
                    error_per_bs[bs_id] = []
                error_per_bs[bs_id].append(error)
                errors.append(error)

        error_info = {}
        error_info['mean_error'] = np.mean(errors)
        error_info['max_error'] = np.max(errors)
        error_info['std_error'] = np.std(errors)

        error_info['bs'] = {}
        for bs_id, errors in error_per_bs.items():
            error_info['bs'][bs_id] = {}
            error_info['bs'][bs_id]['mean_error'] = np.mean(errors)
            error_info['bs'][bs_id]['max_error'] = np.max(errors)
            error_info['bs'][bs_id]['std_error'] = np.std(errors)

        return error_info



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Functionality to handle base station geometry in the lighthouse poistioning system
"""
import math

import numpy as np
import pkg_resources
installed = {pkg.key for pkg in pkg_resources.working_set}
if {'opencv-python-headless'} - installed:
    OPENCV_INSTALLED = False
else:
    import cv2 as cv
    OPENCV_INSTALLED = True


class LighthouseBsGeoEstimator:
    """
    This class is used to estimate the geometry (position and attitude)
    of a lighthouse base station, given angles measured using a lighthouse deck.
    """

    def __init__(self):

        self._estimator_available = True
        if OPENCV_INSTALLED is False:
            self._estimator_available = False

        self._directions = {
            self._hash_sensor_order([2, 0, 1, 3]): math.radians(0),
            self._hash_sensor_order([2, 0, 3, 1]): math.radians(25),
            self._hash_sensor_order([2, 3, 0, 1]): math.radians(65),
            self._hash_sensor_order([3, 2, 0, 1]): math.radians(90),
            self._hash_sensor_order([3, 2, 1, 0]): math.radians(115),
            self._hash_sensor_order([3, 1, 2, 0]): math.radians(155),
            self._hash_sensor_order([1, 3, 2, 0]): math.radians(180),
            self._hash_sensor_order([1, 3, 0, 2]): math.radians(205),
            self._hash_sensor_order([1, 0, 3, 2]): math.radians(245),
            self._hash_sensor_order([0, 1, 3, 2]): math.radians(270),
            self._hash_sensor_order([0, 1, 2, 3]): math.radians(295),
            self._hash_sensor_order([0, 2, 1, 3]): math.radians(335),
        }

        # Sensor distances on the lighthouse deck
        sensor_distance_width = 0.015
        sensor_distance_length = 0.03

        # Sensor positions in world coordinates, open cv style
        self._lighthouse_3d = np.float32(
            [
                [-sensor_distance_width / 2, 0, -sensor_distance_length / 2],
                [sensor_distance_width / 2, 0, -sensor_distance_length / 2],
                [-sensor_distance_width / 2, 0, sensor_distance_length / 2],
                [sensor_distance_width / 2, 0, sensor_distance_length / 2]
            ])

        # Camera matrix
        self._K = np.float64(
            [
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0]
            ])

        self._dist_coef = np.zeros(4)

        # Sanity check maximum pos
        self._sanity_max_pos = 10

    def is_available(self):
        return self._estimator_available

    def estimate_geometry(self, bs_vectors):
        """
        Estimate the full pose of a base station based on angles from the 4 sensors
        on a lighthouse deck. The result is a rotation matrix and position of the
        base station, in the Crazyflie reference frame.

        :param bs_vectors A list of 4 LighthouseBsVector objects specifying vectors to the 4 sensors
        :return rot_bs_in_cf_coord: Rotation matrix of the BS in the CFs coordinate system
        :return pos_bs_in_cf_coord: Position vector of the BS in the CFs coordinate system
        """

        if OPENCV_INSTALLED is False:
            raise Exception('OpenCV is not installed. To use this function,' +
                            'do "pip3 install opencv-python-headless"' +
                            ' and restart the cfclient')

        guess_yaw = self._find_initial_yaw_guess(bs_vectors)
        rvec_guess, tvec_guess = self._convert_yaw_to_open_cv(guess_yaw)
        rw_ocv, tw_ocv = self._estimate_pose_by_pnp(bs_vectors, rvec_guess, tvec_guess)
        rot_bs_in_cf_coord, pos_bs_in_cf_coord = self._opencv_to_cf(rw_ocv, tw_ocv)
        return rot_bs_in_cf_coord, pos_bs_in_cf_coord

    def sanity_check_result(self, pos_bs_in_cf_coord):
        """
        Checks if the estimated geometry is within reasonable bounds. It returns
        true if it seems reasonable or false if it doesn't
        """
        for coord in pos_bs_in_cf_coord:
            if (abs(coord) > self._sanity_max_pos):
                return False
        return True

    def _find_initial_yaw_guess(self, bs_vectors):
        # Assume bs is faicing slighly downwards and fairly horizontal
        # Sort sensors in the order they are hit by the horizontal sweep
        # and use the order to figure out roughly the direction to the
        # base station
        sweeps_x = {
            0: bs_vectors[0].lh_v1_horiz_angle,
            1: bs_vectors[1].lh_v1_horiz_angle,
            2: bs_vectors[2].lh_v1_horiz_angle,
            3: bs_vectors[3].lh_v1_horiz_angle
        }

        ordered_map = {k: v for k, v in sorted(sweeps_x.items(), key=lambda item: item[1])}
        sensor_order = list(ordered_map.keys())

        # The base station is roughly in this direction, in CF (world) coordinates
        return self._directions[self._hash_sensor_order(sensor_order)]

    def _hash_sensor_order(self, order):
        hash = 0
        for i in range(4):
            hash += order[i] * 4 ** i
        return hash

    def _convert_yaw_to_open_cv(self, yaw):
        # Base station height
        bs_h = 2.0
        # Distance to base station along the floor
        bs_fd = 3.0
        # Distance to base station
        bs_dist = math.sqrt(bs_h ** 2 + bs_fd ** 2)
        elevation = math.atan2(bs_h, bs_fd)

        # Initial position of the CF in camera coordinate system, open cv style
        tvec_start = np.array([0, 0, bs_dist])

        # Calculate rotation matrix
        d_c = math.cos(-yaw + math.pi)
        d_s = math.sin(-yaw + math.pi)
        R_rot_y = np.array([
            [d_c, 0.0, d_s],
            [0.0, 1.0, 0.0],
            [-d_s, 0.0, d_c],
        ])

        e_c = math.cos(elevation)
        e_s = math.sin(elevation)
        R_rot_x = np.array([
            [1.0, 0.0, 0.0],
            [0.0, e_c, -e_s],
            [0.0, e_s, e_c],
        ])

        R = np.dot(R_rot_x, R_rot_y)
        rvec_start, _ = cv.Rodrigues(R)

        return rvec_start, tvec_start

    def _estimate_pose_by_pnp(self, bs_vectors, rvec_start, tvec_start):
        # Sensors as seen by the "camera"
        lighthouse_image_projection = np.float32(
            [
                [-math.tan(bs_vectors[0].lh_v1_horiz_angle), -math.tan(bs_vectors[0].lh_v1_vert_angle)],
                [-math.tan(bs_vectors[1].lh_v1_horiz_angle), -math.tan(bs_vectors[1].lh_v1_vert_angle)],
                [-math.tan(bs_vectors[2].lh_v1_horiz_angle), -math.tan(bs_vectors[2].lh_v1_vert_angle)],
                [-math.tan(bs_vectors[3].lh_v1_horiz_angle), -math.tan(bs_vectors[3].lh_v1_vert_angle)]
            ])

        _ret, rvec_est, tvec_est = cv.solvePnP(
            self._lighthouse_3d,
            lighthouse_image_projection,
            self._K,
            self._dist_coef,
            flags=cv.SOLVEPNP_ITERATIVE,
            rvec=rvec_start,
            tvec=tvec_start,
            useExtrinsicGuess=True)

        if not _ret:
            raise Exception('No solution found')

        Rw_ocv, Tw_ocv = self._cam_to_world(rvec_est, tvec_est)
        return Rw_ocv, Tw_ocv

    def _cam_to_world(self, rvec_c, tvec_c):
        R_c, _ = cv.Rodrigues(rvec_c)
        R_w = np.linalg.inv(R_c)
        tvec_w = -np.matmul(R_w, tvec_c)
        return R_w, tvec_w

    def _opencv_to_cf(self, R_cv, t_cv):
        R_opencv_to_cf = np.array([
            [0.0, 0.0, 1.0],
            [-1.0, 0.0, 0.0],
            [0.0, -1.0, 0.0],
        ])

        R_cf_to_opencv = np.array([
            [0.0, -1.0, 0.0],
            [0.0, 0.0, -1.0],
            [1.0, 0.0, 0.0],
        ])

        t_cf = np.dot(R_opencv_to_cf, t_cv)
        R_cf = np.dot(R_opencv_to_cf, np.dot(R_cv, R_cf_to_opencv))

        return R_cf, t_cf



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import unittest

from cflib.localization.lighthouse_bs_vector import LighthouseBsVector
from cflib.localization.lighthouse_sample_matcher import LighthouseSampleMatcher
from cflib.localization.lighthouse_types import LhMeasurement


class TestLighthouseSampleMatcher(unittest.TestCase):
    def setUp(self):
        self.vec0 = LighthouseBsVector(0.0, 0.0)
        self.vec1 = LighthouseBsVector(0.1, 0.1)
        self.vec2 = LighthouseBsVector(0.2, 0.2)
        self.vec3 = LighthouseBsVector(0.3, 0.3)

        self.samples = [
            LhMeasurement(timestamp=1.000, base_station_id=0, angles=self.vec0),
            LhMeasurement(timestamp=1.015, base_station_id=1, angles=self.vec1),
            LhMeasurement(timestamp=1.020, base_station_id=0, angles=self.vec2),
            LhMeasurement(timestamp=1.035, base_station_id=1, angles=self.vec3),
        ]

    def test_that_samples_are_aggregated(self):
        # Fixture

        # Test
        actual = LighthouseSampleMatcher.match(self.samples, max_time_diff=0.010)

        # Assert
        self.assertEqual(1.000, actual[0].timestamp)
        self.assertEqual(1, len(actual[0].angles_calibrated))
        self.assertEqual(self.vec0, actual[0].angles_calibrated[0])

        self.assertEqual(1.015, actual[1].timestamp)
        self.assertEqual(2, len(actual[1].angles_calibrated))
        self.assertEqual(self.vec1, actual[1].angles_calibrated[1])
        self.assertEqual(self.vec2, actual[1].angles_calibrated[0])

        self.assertEqual(1.035, actual[2].timestamp)
        self.assertEqual(1, len(actual[2].angles_calibrated))
        self.assertEqual(self.vec3, actual[2].angles_calibrated[1])

    def test_that_single_bs_samples_are_fitered_out(self):
        # Fixture

        # Test
        actual = LighthouseSampleMatcher.match(self.samples, max_time_diff=0.010, min_nr_of_bs_in_match=2)

        # Assert
        self.assertEqual(1.015, actual[0].timestamp)
        self.assertEqual(2, len(actual[0].angles_calibrated))
        self.assertEqual(self.vec1, actual[0].angles_calibrated[1])
        self.assertEqual(self.vec2, actual[0].angles_calibrated[0])

    def test_that_empty_sample_list_works(self):
        # Fixture

        # Test
        actual = LighthouseSampleMatcher.match([])

        # Assert
        self.assertEqual(0, len(actual))



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
import struct

from .memory_element import MemoryElement

logger = logging.getLogger(__name__)


class AnchorData2:
    """Holds data for one anchor"""

    def __init__(self, position=(0.0, 0.0, 0.0), is_valid=False):
        self.position = position
        self.is_valid = is_valid

    def set_from_mem_data(self, data):
        x, y, z, self.is_valid = struct.unpack('<fff?', data)
        self.position = (x, y, z)


class LocoMemory2(MemoryElement):
    """Memory interface for accessing data from the Loco Positioning system
       version 2"""

    SIZE_OF_FLOAT = 4

    # MAX_NR_OF_ANCHORS should be set to the number of anchors
    # supported by the firmware. Preferably short enough to fit into one packet
    MAX_NR_OF_ANCHORS = 16
    ID_LIST_LEN = 1 + MAX_NR_OF_ANCHORS

    ADR_ID_LIST = 0x0000
    ADR_ACTIVE_ID_LIST = 0x1000
    ADR_ANCHOR_BASE = 0x2000

    ANCHOR_PAGE_SIZE = 0x0100
    PAGE_LEN = (3 * SIZE_OF_FLOAT) + 1

    def __init__(self, id, type, size, mem_handler):
        super(LocoMemory2, self).__init__(id=id, type=type, size=size,
                                          mem_handler=mem_handler)
        self._update_ids_finished_cb = None
        self._update_active_ids_finished_cb = None
        self._update_data_finished_cb = None
        self._currently_fetching_index = -1

        self.anchor_ids = []
        self.active_anchor_ids = []
        self.anchor_data = {}
        self.nr_of_anchors = 0
        self.ids_valid = False
        self.active_ids_valid = False
        self.data_valid = False

    def new_data(self, mem, addr, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            if addr == LocoMemory2.ADR_ID_LIST:
                self._handle_id_list_data(data)
            elif addr == LocoMemory2.ADR_ACTIVE_ID_LIST:
                self._handle_active_id_list_data(data)
            else:
                id = int((addr - LocoMemory2.ADR_ANCHOR_BASE) /
                         LocoMemory2.ANCHOR_PAGE_SIZE)
                self._handle_anchor_data(id, data)

    def update_id_list(self, update_ids_finished_cb):
        """Request an update of the id list"""
        if not self._update_ids_finished_cb:
            self._update_ids_finished_cb = update_ids_finished_cb
            self.anchor_ids = []
            self.active_anchor_ids = []
            self.anchor_data = {}

            self.nr_of_anchors = 0
            self.ids_valid = False
            self.data_valid = False

            logger.debug('Updating ids of memory {}'.format(self.id))

            # Start reading the header
            self.mem_handler.read(self, LocoMemory2.ADR_ID_LIST,
                                  LocoMemory2.ID_LIST_LEN)

    def update_active_id_list(self, update_active_ids_finished_cb):
        """Request an update of the active id list"""
        if not self._update_active_ids_finished_cb:
            self._update_active_ids_finished_cb = update_active_ids_finished_cb
            self.active_anchor_ids = []

            self.active_ids_valid = False

            logger.debug('Updating active ids of memory {}'.format(self.id))

            # Start reading the header
            self.mem_handler.read(self, LocoMemory2.ADR_ACTIVE_ID_LIST,
                                  LocoMemory2.ID_LIST_LEN)

    def update_data(self, update_data_finished_cb):
        """Request an update of the anchor data"""
        if not self._update_data_finished_cb and self.nr_of_anchors > 0:
            self._update_data_finished_cb = update_data_finished_cb
            self.anchor_data = {}

            self.data_valid = False
            self._nr_of_anchors_to_fetch = self.nr_of_anchors

            logger.debug('Updating anchor data of memory {}'.format(self.id))

            # Start reading the first anchor
            self._currently_fetching_index = 0
            self._request_page(self.anchor_ids[self._currently_fetching_index])

    def disconnect(self):
        self._update_ids_finished_cb = None
        self._update_data_finished_cb = None

    def _handle_id_list_data(self, data):
        self.nr_of_anchors = data[0]
        for i in range(self.nr_of_anchors):
            self.anchor_ids.append(data[1 + i])
        self.ids_valid = True

        if self._update_ids_finished_cb:
            self._update_ids_finished_cb(self)
            self._update_ids_finished_cb = None

    def _handle_active_id_list_data(self, data):
        count = data[0]
        for i in range(count):
            self.active_anchor_ids.append(data[1 + i])
        self.active_ids_valid = True

        if self._update_active_ids_finished_cb:
            self._update_active_ids_finished_cb(self)
            self._update_active_ids_finished_cb = None

    def _handle_anchor_data(self, id, data):
        anchor = AnchorData2()
        anchor.set_from_mem_data(data)
        self.anchor_data[id] = anchor

        self._currently_fetching_index += 1
        if self._currently_fetching_index < self.nr_of_anchors:
            self._request_page(self.anchor_ids[self._currently_fetching_index])
        else:
            self.data_valid = True
            if self._update_data_finished_cb:
                self._update_data_finished_cb(self)
                self._update_data_finished_cb = None

    def _request_page(self, page):
        addr = LocoMemory2.ADR_ANCHOR_BASE + \
            LocoMemory2.ANCHOR_PAGE_SIZE * page
        self.mem_handler.read(self, addr, LocoMemory2.PAGE_LEN)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018-2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Used for sending high level setpoints to the Crazyflie
"""
import struct

from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort

__author__ = 'Bitcraze AB'
__all__ = ['HighLevelCommander']


class HighLevelCommander():
    """
    Used for sending high level setpoints to the Crazyflie
    """

    COMMAND_SET_GROUP_MASK = 0
    COMMAND_STOP = 3
    COMMAND_GO_TO = 4
    COMMAND_START_TRAJECTORY = 5
    COMMAND_DEFINE_TRAJECTORY = 6
    COMMAND_TAKEOFF_2 = 7
    COMMAND_LAND_2 = 8

    ALL_GROUPS = 0

    TRAJECTORY_LOCATION_MEM = 1

    TRAJECTORY_TYPE_POLY4D = 0
    TRAJECTORY_TYPE_POLY4D_COMPRESSED = 1

    def __init__(self, crazyflie=None):
        """
        Initialize the object.
        """
        self._cf = crazyflie

    def set_group_mask(self, group_mask=ALL_GROUPS):
        """
        Set the group mask that the Crazyflie belongs to

        :param group_mask: Mask for which groups this CF belongs to
        """
        self._send_packet(struct.pack('<BB',
                                      self.COMMAND_SET_GROUP_MASK,
                                      group_mask))

    def takeoff(self, absolute_height_m, duration_s, group_mask=ALL_GROUPS,
                yaw=0.0):
        """
        vertical takeoff from current x-y position to given height

        :param absolute_height_m: Absolute (m)
        :param duration_s: Time it should take until target height is
                           reached (s)
        :param group_mask: Mask for which CFs this should apply to
        :param yaw: Yaw (rad). Use current yaw if set to None.
        """
        target_yaw = yaw
        useCurrentYaw = False
        if yaw is None:
            target_yaw = 0.0
            useCurrentYaw = True

        self._send_packet(struct.pack('<BBff?f',
                                      self.COMMAND_TAKEOFF_2,
                                      group_mask,
                                      absolute_height_m,
                                      target_yaw,
                                      useCurrentYaw,
                                      duration_s))

    def land(self, absolute_height_m, duration_s, group_mask=ALL_GROUPS,
             yaw=0.0):
        """
        vertical land from current x-y position to given height

        :param absolute_height_m: Absolute (m)
        :param duration_s: Time it should take until target height is
                           reached (s)
        :param group_mask: Mask for which CFs this should apply to
        :param yaw: Yaw (rad). Use current yaw if set to None.
        """
        target_yaw = yaw
        useCurrentYaw = False
        if yaw is None:
            target_yaw = 0.0
            useCurrentYaw = True

        self._send_packet(struct.pack('<BBff?f',
                                      self.COMMAND_LAND_2,
                                      group_mask,
                                      absolute_height_m,
                                      target_yaw,
                                      useCurrentYaw,
                                      duration_s))

    def stop(self, group_mask=ALL_GROUPS):
        """
        stops the current trajectory (turns off the motors)

        :param group_mask: Mask for which CFs this should apply to
        :return:
        """
        self._send_packet(struct.pack('<BB',
                                      self.COMMAND_STOP,
                                      group_mask))

    def go_to(self, x, y, z, yaw, duration_s, relative=False,
              group_mask=ALL_GROUPS):
        """
        Go to an absolute or relative position

        :param x: X (m)
        :param y: Y (m)
        :param z: Z (m)
        :param yaw: Yaw (radians)
        :param duration_s: Time it should take to reach the position (s)
        :param relative: True if x, y, z is relative to the current position
        :param group_mask: Mask for which CFs this should apply to
        """
        self._send_packet(struct.pack('<BBBfffff',
                                      self.COMMAND_GO_TO,
                                      group_mask,
                                      relative,
                                      x, y, z,
                                      yaw,
                                      duration_s))

    def start_trajectory(self, trajectory_id, time_scale=1.0, relative=False,
                         reversed=False, group_mask=ALL_GROUPS):
        """
        starts executing a specified trajectory

        :param trajectory_id: Id of the trajectory (previously defined by
               define_trajectory)
        :param time_scale: Time factor; 1.0 = original speed;
                                        >1.0: slower;
                                        <1.0: faster
        :param relative: Set to True, if trajectory should be shifted to
               current setpoint
        :param reversed: Set to True, if trajectory should be executed in
               reverse
        :param group_mask: Mask for which CFs this should apply to
        :return:
        """
        self._send_packet(struct.pack('<BBBBBf',
                                      self.COMMAND_START_TRAJECTORY,
                                      group_mask,
                                      relative,
                                      reversed,
                                      trajectory_id,
                                      time_scale))

    def define_trajectory(self, trajectory_id, offset, n_pieces, type=TRAJECTORY_TYPE_POLY4D):
        """
        Define a trajectory that has previously been uploaded to memory.

        :param trajectory_id: The id of the trajectory
        :param offset: Offset in uploaded memory
        :param n_pieces: Nr of pieces in the trajectory
        :param type: The type of trajectory data; TRAJECTORY_TYPE_POLY4D or TRAJECTORY_TYPE_POLY4D_COMPRESSED
        :return:
        """
        self._send_packet(struct.pack('<BBBBIB',
                                      self.COMMAND_DEFINE_TRAJECTORY,
                                      trajectory_id,
                                      self.TRAJECTORY_LOCATION_MEM,
                                      type,
                                      offset,
                                      n_pieces))

    def _send_packet(self, data):
        pk = CRTPPacket()
        pk.port = CRTPPort.SETPOINT_HL
        pk.data = data
        self._cf.send_packet(pk)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from test.localization.lighthouse_fixtures import LighthouseFixtures
from test.localization.lighthouse_test_base import LighthouseTestBase

import numpy as np

from cflib.localization.lighthouse_initial_estimator import LighthouseInitialEstimator
from cflib.localization.lighthouse_types import LhCfPoseSample
from cflib.localization.lighthouse_types import LhDeck4SensorPositions
from cflib.localization.lighthouse_types import Pose


class TestLighthouseInitialEstimator(LighthouseTestBase):
    def setUp(self):
        self.fixtures = LighthouseFixtures()

    def test_that_one_bs_pose_raises_exception(self):
        # Fixture
        # CF_ORIGIN is used in the first sample and will define the global reference frame
        bs_id = 3
        samples = [
            LhCfPoseSample(angles_calibrated={bs_id: self.fixtures.angles_cf_origin_bs0}),
        ]

        # Test
        # Assert
        with self.assertRaises(Exception):
            LighthouseInitialEstimator.estimate(samples, LhDeck4SensorPositions.positions)

    def test_that_two_bs_poses_in_same_sample_are_found(self):
        # Fixture
        # CF_ORIGIN is used in the first sample and will define the global reference frame
        bs_id0 = 3
        bs_id1 = 1
        samples = [
            LhCfPoseSample(angles_calibrated={
                bs_id0: self.fixtures.angles_cf_origin_bs0,
                bs_id1: self.fixtures.angles_cf_origin_bs1,
            }),
        ]

        # Test
        actual, cleaned_samples = LighthouseInitialEstimator.estimate(samples, LhDeck4SensorPositions.positions)

        # Assert
        self.assertPosesAlmostEqual(self.fixtures.BS0_POSE, actual.bs_poses[bs_id0], places=3)
        self.assertPosesAlmostEqual(self.fixtures.BS1_POSE, actual.bs_poses[bs_id1], places=3)

    def test_that_linked_bs_poses_in_multiple_samples_are_found(self):
        # Fixture
        # CF_ORIGIN is used in the first sample and will define the global reference frame
        bs_id0 = 7
        bs_id1 = 1
        bs_id2 = 5
        bs_id3 = 0
        samples = [
            LhCfPoseSample(angles_calibrated={
                bs_id0: self.fixtures.angles_cf_origin_bs0,
                bs_id1: self.fixtures.angles_cf_origin_bs1,
            }),
            LhCfPoseSample(angles_calibrated={
                bs_id1: self.fixtures.angles_cf1_bs1,
                bs_id2: self.fixtures.angles_cf1_bs2,
            }),
            LhCfPoseSample(angles_calibrated={
                bs_id2: self.fixtures.angles_cf2_bs2,
                bs_id3: self.fixtures.angles_cf2_bs3,
            }),
        ]

        # Test
        actual, cleaned_samples = LighthouseInitialEstimator.estimate(samples, LhDeck4SensorPositions.positions)

        # Assert
        self.assertPosesAlmostEqual(self.fixtures.BS0_POSE, actual.bs_poses[bs_id0], places=3)
        self.assertPosesAlmostEqual(self.fixtures.BS1_POSE, actual.bs_poses[bs_id1], places=3)
        self.assertPosesAlmostEqual(self.fixtures.BS2_POSE, actual.bs_poses[bs_id2], places=3)
        self.assertPosesAlmostEqual(self.fixtures.BS3_POSE, actual.bs_poses[bs_id3], places=3)

    def test_that_cf_poses_are_estimated(self):
        # Fixture
        # CF_ORIGIN is used in the first sample and will define the global reference frame
        bs_id0 = 7
        bs_id1 = 1
        bs_id2 = 5
        bs_id3 = 0
        samples = [
            LhCfPoseSample(angles_calibrated={
                bs_id0: self.fixtures.angles_cf_origin_bs0,
                bs_id1: self.fixtures.angles_cf_origin_bs1,
            }),
            LhCfPoseSample(angles_calibrated={
                bs_id1: self.fixtures.angles_cf1_bs1,
                bs_id2: self.fixtures.angles_cf1_bs2,
            }),
            LhCfPoseSample(angles_calibrated={
                bs_id2: self.fixtures.angles_cf2_bs2,
                bs_id3: self.fixtures.angles_cf2_bs3,
            }),
        ]

        # Test
        actual, cleaned_samples = LighthouseInitialEstimator.estimate(samples, LhDeck4SensorPositions.positions)

        # Assert
        self.assertPosesAlmostEqual(self.fixtures.CF_ORIGIN_POSE, actual.cf_poses[0], places=3)
        self.assertPosesAlmostEqual(self.fixtures.CF1_POSE, actual.cf_poses[1], places=3)
        self.assertPosesAlmostEqual(self.fixtures.CF2_POSE, actual.cf_poses[2], places=3)

    def test_that_the_global_ref_frame_is_used(self):
        # Fixture
        # CF2 is used in the first sample and will define the global reference frame
        bs_id0 = 3
        bs_id1 = 1
        bs_id2 = 2
        samples = [
            LhCfPoseSample(angles_calibrated={
                bs_id0: self.fixtures.angles_cf2_bs0,
                bs_id1: self.fixtures.angles_cf2_bs1,
            }),
            LhCfPoseSample(angles_calibrated={
                bs_id1: self.fixtures.angles_cf1_bs1,
                bs_id2: self.fixtures.angles_cf1_bs2,
            }),
        ]

        # Test
        actual, cleaned_samples = LighthouseInitialEstimator.estimate(samples, LhDeck4SensorPositions.positions)

        # Assert
        self.assertPosesAlmostEqual(
            Pose.from_rot_vec(R_vec=(0.0, 0.0, -np.pi / 2), t_vec=(1.0, 3.0, 3.0)), actual.bs_poses[bs_id0], places=3)
        self.assertPosesAlmostEqual(
            Pose.from_rot_vec(R_vec=(0.0, 0.0, 0.0), t_vec=(-2.0, 1.0, 3.0)), actual.bs_poses[bs_id1], places=3)
        self.assertPosesAlmostEqual(
            Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi), t_vec=(2.0, 1.0, 3.0)), actual.bs_poses[bs_id2], places=3)

    def test_that_raises_for_isolated_bs(self):
        # Fixture
        bs_id0 = 3
        bs_id1 = 1
        bs_id2 = 2
        bs_id3 = 4
        samples = [
            LhCfPoseSample(angles_calibrated={
                bs_id0: self.fixtures.angles_cf_origin_bs0,
                bs_id1: self.fixtures.angles_cf_origin_bs1,
            }),
            LhCfPoseSample(angles_calibrated={
                bs_id2: self.fixtures.angles_cf1_bs2,
                bs_id3: self.fixtures.angles_cf2_bs2,
            }),
        ]

        # Test
        # Assert
        with self.assertRaises(Exception):
            LighthouseInitialEstimator.estimate(samples, LhDeck4SensorPositions.positions)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2022 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
""" CRTP CPX Driver. Tunnel CRTP over CPX to the Crazyflie STM32 """
import logging
import queue
import re
import struct
import threading
from urllib.parse import urlparse

from .crtpdriver import CRTPDriver
from .crtpstack import CRTPPacket
from .exceptions import WrongUriType
from cflib.cpx import CPX
from cflib.cpx import CPXFunction
from cflib.cpx import CPXPacket
from cflib.cpx import CPXTarget
from cflib.cpx.transports import SocketTransport
logger = logging.getLogger(__name__)

__author__ = 'Bitcraze AB'
__all__ = ['TcpDriver']


class TcpDriver(CRTPDriver):

    def __init__(self):
        self.needs_resending = False

    def connect(self, uri, linkQualityCallback, linkErrorCallback):
        if not re.search('^tcp://', uri):
            raise WrongUriType('Not an UDP URI')

        parse = urlparse(uri.split(' ')[0])

        self.in_queue = queue.Queue()

        self.cpx = CPX(SocketTransport(parse.hostname, parse.port))

        self._thread = _CPXReceiveThread(self.cpx, self.in_queue,
                                         linkErrorCallback)
        self._thread.start()

        self.cpx.sendPacket(CPXPacket(destination=CPXTarget.STM32,
                                      function=CPXFunction.SYSTEM,
                                      data=[0x21, 0x01]))

    def receive_packet(self, time=0):
        if time == 0:
            try:
                return self.in_queue.get(False)
            except queue.Empty:
                return None
        elif time < 0:
            try:
                return self.in_queue.get(True)
            except queue.Empty:
                return None
        else:
            try:
                return self.in_queue.get(True, time)
            except queue.Empty:
                return None

    def send_packet(self, pk):
        raw = (pk.header,) + struct.unpack('B' * len(pk.data), pk.data)
        self.cpx.sendPacket(CPXPacket(destination=CPXTarget.STM32,
                                      function=CPXFunction.CRTP,
                                      data=raw))

    def close(self):
        """ Close the link. """
        # Stop the comm thread
        self._thread.stop()

        # Close the socket
        try:
            self.cpx.close()
            self.cpx = None

        except Exception as e:
            print(e)
            logger.error('Could not close {}'.format(e))
            pass
        print('Driver closed')
        self.cpx = None

    def get_name(self):
        return 'cpx'

    def scan_interface(self, address):
        return []

# Transmit/receive thread


class _CPXReceiveThread(threading.Thread):
    """
    Radio link receiver thread used to read data from the
    Socket. """

    def __init__(self, cpx, inQueue, link_error_callback):
        """ Create the object """
        threading.Thread.__init__(self)
        self._cpx = cpx
        self.in_queue = inQueue
        self.sp = False
        self.link_error_callback = link_error_callback

    def stop(self):
        """ Stop the thread """
        self.sp = True
        try:
            self.join()
        except Exception:
            pass

    def run(self):
        """ Run the receiver thread """

        while (True):
            if (self.sp):
                break
            try:
                # Block until a packet is available though the socket
                # CPX receive will only return full packets
                cpxPacket = self._cpx.receivePacket(CPXFunction.CRTP, timeout=0.1)
                data = struct.unpack('B' * len(cpxPacket.data), cpxPacket.data)
                if len(data) > 0:
                    pk = CRTPPacket(data[0],
                                    list(data[1:]))
                    self.in_queue.put(pk)
            except queue.Empty:
                pass  # This is ok
            except Exception as e:
                import traceback

                self.link_error_callback(
                    'Error communicating with the Crazyflie\n'
                    'Exception:%s\n\n%s' % (e,
                                            traceback.format_exc()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.
"""
CRTP packet and ports.
"""
import logging

__author__ = 'Bitcraze AB'
__all__ = ['CRTPPort', 'CRTPPacket']

logger = logging.getLogger(__name__)


class CRTPPort:
    """
    Lists the available ports for the CRTP.
    """
    CONSOLE = 0x00
    PARAM = 0x02
    COMMANDER = 0x03
    MEM = 0x04
    LOGGING = 0x05
    LOCALIZATION = 0x06
    COMMANDER_GENERIC = 0x07
    SETPOINT_HL = 0x08
    PLATFORM = 0x0D
    LINKCTRL = 0x0F
    ALL = 0xFF


class CRTPPacket(object):
    """
    A packet that can be sent via the CRTP.
    """

    # The max size of a CRTP packet payload
    MAX_DATA_SIZE = 30

    def __init__(self, header=0, data=None):
        """
        Create an empty packet with default values.
        """
        self.size = 0
        self._data = bytearray()
        # The two bits in position 3 and 4 needs to be set for legacy
        # support of the bootloader
        self.header = header | 0x3 << 2
        self._port = (header & 0xF0) >> 4
        self._channel = header & 0x03
        if data:
            self._set_data(data)

    def _get_channel(self):
        """Get the packet channel"""
        return self._channel

    def _set_channel(self, channel):
        """Set the packet channel"""
        self._channel = channel
        self._update_header()

    def _get_port(self):
        """Get the packet port"""
        return self._port

    def _set_port(self, port):
        """Set the packet port"""
        self._port = port
        self._update_header()

    def get_header(self):
        """Get the header"""
        self._update_header()
        return self.header

    def set_header(self, port, channel):
        """
        Set the port and channel for this packet.
        """
        self._port = port
        self.channel = channel
        self._update_header()

    def _update_header(self):
        """Update the header with the port/channel values"""
        # The two bits in position 3 and 4 needs to be set for legacy
        # support of the bootloader
        self.header = ((self._port & 0x0f) << 4 | 3 << 2 |
                       (self.channel & 0x03))

    # Some python madness to access different format of the data
    def _get_data(self):
        """Get the packet data"""
        return self._data

    def _set_data(self, data):
        """Set the packet data"""
        if type(data) == bytearray:
            self._data = data
        elif type(data) == str:
            self._data = bytearray(data.encode('ISO-8859-1'))
        elif type(data) == list or type(data) == tuple:
            self._data = bytearray(data)
        elif type(data) == bytes:
            self._data = bytearray(data)
        else:
            raise Exception('Data must be bytearray, string, list or tuple,'
                            ' not {}'.format(type(data)))

    def _get_data_l(self):
        """Get the data in the packet as a list"""
        return list(self._get_data_t())

    def _get_data_t(self):
        """Get the data in the packet as a tuple"""
        return tuple(self._data)

    def __str__(self):
        """Get a string representation of the packet"""
        return '{}:{} {}'.format(self._port, self.channel, self.datat)

    def get_data_size(self):
        return len(self._data)

    def available_data_size(self):
        return self.MAX_DATA_SIZE - self.get_data_size()

    def is_data_size_valid(self):
        return self.available_data_size() >= 0

    data = property(_get_data, _set_data)
    datal = property(_get_data_l, _set_data)
    datat = property(_get_data_t, _set_data)
    datas = property(_get_data, _set_data)
    port = property(_get_port, _set_port)
    channel = property(_get_channel, _set_channel)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021-2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.optimize

from cflib.localization.lighthouse_types import Pose


class LighthouseSystemAligner:
    """This class is used to align a lighthouse system to a few sampled positions"""
    @classmethod
    def align(cls, origin: npt.ArrayLike, x_axis: list[npt.ArrayLike], xy_plane: list[npt.ArrayLike],
              bs_poses: dict[int, Pose]) -> tuple[dict[int, Pose], Pose]:
        """
        Align a coordinate system with the physical world. Finds the transform from the
        current reference frame to one that is aligned with measured positions, and transforms base station
        poses to the new coordinate system.

        :param origin: The position of the desired origin in the current reference frame
        :param x_axis: One or more positions on the desired positive X-axis (X>0, Y=Z=0) in the current
                       reference frame
        :param x_axis: One or more positions in the desired XY-plane (Z=0) in the current reference frame
        :param bs_poses: a dictionary with the base station poses in the current reference frame
        :return: a dictionary with the base station poses in the desired reference frame and the transformation
        """
        raw_transformation = cls._find_transformation(origin, x_axis, xy_plane)
        transformation = cls._de_flip_transformation(raw_transformation, x_axis, bs_poses)

        result: dict[int, Pose] = {}
        for bs_id, pose in bs_poses.items():
            result[bs_id] = transformation.rotate_translate_pose(pose)

        return result, transformation

    @classmethod
    def _find_transformation(cls, origin: npt.ArrayLike, x_axis: list[npt.ArrayLike],
                             xy_plane: list[npt.ArrayLike]) -> Pose:
        """
        Finds the transformation from the current reference frame to a desired reference frame based on measured
        positions of the desired reference frame.

        :param origin: The position of the desired origin in the current reference frame
        :param x_axis: One or more positions on the desired positive X-axis (X>0, Y=Z=0) in the current
                       reference frame
        :param x_axis: One or more positions in the desired XY-plane (Z=0) in the current reference frame
        :return: The transformation from the current reference frame to the desired reference frame. Note: the
                 solution may be flipped.
        """
        args = (origin, x_axis, xy_plane)

        x0 = np.zeros((6))

        result = scipy.optimize.least_squares(cls._calc_residual,
                                              x0, verbose=0,
                                              jac_sparsity=None,
                                              x_scale='jac',
                                              ftol=1e-8,
                                              method='trf',
                                              max_nfev=10,
                                              args=args)
        return cls._Pose_from_params(result.x)

    @classmethod
    def _calc_residual(cls, params, origin: npt.ArrayLike, x_axis: list[npt.ArrayLike], xy_plane: list[npt.ArrayLike]):
        transform = cls._Pose_from_params(params)

        origin_diff = transform.rotate_translate(origin)
        x_axis_diff = map(lambda x: transform.rotate_translate(x), x_axis)
        xy_plane_diff = map(lambda x: transform.rotate_translate(x), xy_plane)

        residual_origin = origin_diff

        # Points on X-axis: ignore X
        x_axis_residual = list(map(lambda x: x[1:3], x_axis_diff))

        # Points in the XY-plane: ignore X and Y
        xy_plane_residual = list(map(lambda x: x[2], xy_plane_diff))

        residual = np.concatenate((np.ravel(residual_origin), np.ravel(x_axis_residual), np.ravel(xy_plane_residual)))
        return residual

    @classmethod
    def _Pose_from_params(cls, params: npt.ArrayLike) -> Pose:
        return Pose.from_rot_vec(R_vec=params[:3], t_vec=params[3:])

    @classmethod
    def _de_flip_transformation(cls, raw_transformation: Pose, x_axis: list[npt.ArrayLike],
                                bs_poses: dict[int, Pose]) -> Pose:
        """
        Investigats a transformation and flips it if needed. This method assumes that
        1. all base stations are at Z>0
        2. x_axis samples are taken at X>0
        """
        transformation = raw_transformation

        # X-axis poses should be on the positivie X-axis, check that the "mean" of the x-axis points ends up at X>0
        x_axis_mean = np.mean(x_axis, axis=0)
        if raw_transformation.rotate_translate(x_axis_mean)[0] < 0.0:
            flip_around_z_axis = Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi))
            transformation = flip_around_z_axis.rotate_translate_pose(transformation)

        # Base station poses should be above the floor, check the first one
        bs_pose = list(bs_poses.values())[0]
        if raw_transformation.rotate_translate(bs_pose.translation)[2] < 0.0:
            flip_around_x_axis = Pose.from_rot_vec(R_vec=(np.pi, 0.0, 0.0))
            transformation = flip_around_x_axis.rotate_translate_pose(transformation)

        return transformation



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from test.localization.lighthouse_fixtures import LighthouseFixtures
from test.localization.lighthouse_test_base import LighthouseTestBase

from cflib.localization.lighthouse_geometry_solver import LighthouseGeometrySolver
from cflib.localization.lighthouse_initial_estimator import LighthouseInitialEstimator
from cflib.localization.lighthouse_types import LhCfPoseSample
from cflib.localization.lighthouse_types import LhDeck4SensorPositions


class TestLighthouseGeometrySolver(LighthouseTestBase):
    def setUp(self):
        self.fixtures = LighthouseFixtures()

    def test_that_two_bs_poses_in_one_sample_are_estimated(self):
        # Fixture
        # CF_ORIGIN is used in the first sample and will define the global reference frame
        bs_id0 = 3
        bs_id1 = 1
        matched_samples = [
            LhCfPoseSample(angles_calibrated={
                bs_id0: self.fixtures.angles_cf_origin_bs0,
                bs_id1: self.fixtures.angles_cf_origin_bs1,
            }),
        ]

        initial_guess, cleaned_matched_samples = LighthouseInitialEstimator.estimate(matched_samples,
                                                                                     LhDeck4SensorPositions.positions)

        # Test
        actual = LighthouseGeometrySolver.solve(
            initial_guess, cleaned_matched_samples, LhDeck4SensorPositions.positions)

        # Assert
        bs_poses = actual.bs_poses
        self.assertPosesAlmostEqual(self.fixtures.BS0_POSE, bs_poses[bs_id0], places=3)
        self.assertPosesAlmostEqual(self.fixtures.BS1_POSE, bs_poses[bs_id1], places=3)

    def test_that_linked_bs_poses_in_multiple_samples_are_estimated(self):
        # Fixture
        # CF_ORIGIN is used in the first sample and will define the global reference frame
        bs_id0 = 7
        bs_id1 = 2
        bs_id2 = 9
        bs_id3 = 3
        matched_samples = [
            LhCfPoseSample(angles_calibrated={
                bs_id0: self.fixtures.angles_cf_origin_bs0,
                bs_id1: self.fixtures.angles_cf_origin_bs1,
            }),
            LhCfPoseSample(angles_calibrated={
                bs_id1: self.fixtures.angles_cf1_bs1,
                bs_id2: self.fixtures.angles_cf1_bs2,
            }),
            LhCfPoseSample(angles_calibrated={
                bs_id2: self.fixtures.angles_cf2_bs2,
                bs_id3: self.fixtures.angles_cf2_bs3,
            }),
        ]

        initial_guess, cleaned_matched_samples = LighthouseInitialEstimator.estimate(matched_samples,
                                                                                     LhDeck4SensorPositions.positions)

        # Test
        actual = LighthouseGeometrySolver.solve(
            initial_guess, cleaned_matched_samples, LhDeck4SensorPositions.positions)

        # Assert
        bs_poses = actual.bs_poses
        self.assertPosesAlmostEqual(self.fixtures.BS0_POSE, bs_poses[bs_id0], places=3)
        self.assertPosesAlmostEqual(self.fixtures.BS1_POSE, bs_poses[bs_id1], places=3)
        self.assertPosesAlmostEqual(self.fixtures.BS2_POSE, bs_poses[bs_id2], places=3)
        self.assertPosesAlmostEqual(self.fixtures.BS3_POSE, bs_poses[bs_id3], places=3)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from test.localization.lighthouse_test_base import LighthouseTestBase

import numpy as np

from cflib.localization.lighthouse_types import Pose


class TestLighthouseTypes(LighthouseTestBase):
    def setUp(self):
        pass

    def test_default_matrix_constructor(self):
        # Fixture
        # Test
        actual = Pose()

        # Assert
        self.assertEqual(0.0, np.linalg.norm(actual.translation))
        self.assertEqual(0.0, np.linalg.norm(actual.rot_matrix - np.identity(3)))

    def test_default_rot_vec_constructor(self):
        # Fixture
        # Test
        actual = Pose.from_rot_vec()

        # Assert
        self.assertEqual(0.0, np.linalg.norm(actual.translation))
        self.assertEqual(0.0, np.linalg.norm(actual.rot_matrix - np.identity(3)))

    def test_rotate_translate(self):
        # Fixture
        transform = Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi / 2), t_vec=(1.0, 0.0, 0.0))
        point = (2.0, 0.0, 0.0)

        # Test
        actual = transform.rotate_translate(point)

        # Assert
        self.assertAlmostEqual(1.0, actual[0])
        self.assertAlmostEqual(2.0, actual[1])
        self.assertAlmostEqual(0.0, actual[2])

    def test_rotate_translate_and_back(self):
        # Fixture
        transform = Pose.from_rot_vec(R_vec=(1.0, 2.0, 3.0), t_vec=(0.1, 0.2, 0.3))
        expected = (2.0, 3.0, 4.0)

        # Test
        actual = transform.inv_rotate_translate(transform.rotate_translate(expected))

        # Assert
        self.assertAlmostEqual(expected[0], actual[0])
        self.assertAlmostEqual(expected[1], actual[1])
        self.assertAlmostEqual(expected[2], actual[2])

    def test_rotate_translate_pose(self):
        # Fixture
        transform = Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi / 2), t_vec=(1.0, 0.0, 0.0))
        pose = Pose(t_vec=(2.0, 0.0, 0.0))
        expected = Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi / 2), t_vec=(1.0, 2.0, 0.0))

        # Test
        actual = transform.rotate_translate_pose(pose)

        # Assert
        self.assertPosesAlmostEqual(expected, actual)

    def test_rotate_translate_pose_and_back(self):
        # Fixture
        transform = Pose.from_rot_vec(R_vec=(1.0, 2.0, 3.0), t_vec=(0.1, 0.2, 0.3))
        expected = Pose(t_vec=(2.0, 3.0, 4.0))

        # Test
        actual = transform.inv_rotate_translate_pose(transform.rotate_translate_pose(expected))

        # Assert
        self.assertPosesAlmostEqual(expected, actual)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import struct


# Code from davidejones at https://gamedev.stackexchange.com/a/28756
def fp16_to_float(float16):
    s = int((float16 >> 15) & 0x00000001)    # sign
    e = int((float16 >> 10) & 0x0000001f)    # exponent
    f = int(float16 & 0x000003ff)            # fraction

    if e == 0:
        if f == 0:
            return int(s << 31)
        else:
            while not (f & 0x00000400):
                f <<= 1
                e -= 1
            e += 1
            f &= ~0x00000400
            # print(s,e,f)
    elif e == 31:
        if f == 0:
            return int((s << 31) | 0x7f800000)
        else:
            return int((s << 31) | 0x7f800000 | (f << 13))

    e += 127 - 15
    f <<= 13
    result = int((s << 31) | (e << 23) | f)
    return struct.unpack('f', struct.pack('I', result))[0]



#!/usr/bin/env python3
from pathlib import Path

from setuptools import find_packages
from setuptools import setup
# read the contents of README.md file fo use in pypi description
directory = Path(__file__).parent
long_description = (directory / 'README.md').read_text()

setup(
    name='cflib',
    version='0.1.20.1',
    packages=find_packages(exclude=['examples', 'test']),

    description='Crazyflie python driver',
    url='https://github.com/bitcraze/crazyflie-lib-python',

    long_description=long_description,
    long_description_content_type='text/markdown',

    author='Bitcraze and contributors',
    author_email='contact@bitcraze.io',
    license='GPLv3',

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ],

    keywords='driver crazyflie quadcopter',

    install_requires=[
        'pyusb>=1.0.0b2',
        'libusb-package~=1.0',
        'scipy~=1.7',
        'numpy>=1.20,<1.25',
    ],

    # $ pip install -e .[dev]
    extras_require={
        'dev': [
            'pre-commit'
        ],
    },
)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
This class is used to turn the power of the Crazyflie on and off via
a Crazyradio.
"""
import time

import cflib.crtp
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.radiodriver import RadioManager


class PowerSwitch:
    BOOTLOADER_CMD_ALLOFF = 0x01
    BOOTLOADER_CMD_SYSOFF = 0x02
    BOOTLOADER_CMD_SYSON = 0x03
    BOOTLOADER_CMD_RESET_INIT = 0xFF
    BOOTLOADER_CMD_RESET = 0xF0

    def __init__(self, uri):
        self.uri = uri
        uri_augmented = uri+'?safelink=0&autoping=0&ackfilter=0'
        self.link = cflib.crtp.get_link_driver(uri_augmented)
        # Switch to legacy mode, if uri options are not available or old Python backend is used
        if not self.link or self.link.get_name() == 'radio':
            uri_parts = cflib.crtp.RadioDriver.parse_uri(uri)
            self.devid = uri_parts[0]
            self.channel = uri_parts[1]
            self.datarate = uri_parts[2]
            self.address = uri_parts[3]
            if self.link:
                self.link.close()
                self.link = None

    def platform_power_down(self):
        """ Power down the platform, both NRF and STM MCUs.
            Same as turning off the platform with the power button."""
        self._send(self.BOOTLOADER_CMD_ALLOFF)

    def stm_power_down(self):
        """ Power down the STM MCU, the NRF will still be powered and handle
            basic radio communication. The STM can be restarted again remotely.
            Note: the power to expansion decks is also turned off. """
        self._send(self.BOOTLOADER_CMD_SYSOFF)

    def stm_power_up(self):
        """ Power up (boot) the STM MCU and decks."""
        self._send(self.BOOTLOADER_CMD_SYSON)

    def stm_power_cycle(self):
        """ Restart the STM MCU by powering it off and on.
            Expansion decks will also be rebooted."""
        self.stm_power_down()
        time.sleep(1)
        self.stm_power_up()

    def reboot_to_fw(self):
        """Reboot the platform and start in firmware mode"""
        self._send(self.BOOTLOADER_CMD_RESET_INIT)
        self._send(self.BOOTLOADER_CMD_RESET, [1])

    def reboot_to_bootloader(self):
        """Reboot the platform and start the bootloader"""
        self._send(self.BOOTLOADER_CMD_RESET_INIT)
        self._send(self.BOOTLOADER_CMD_RESET, [0])

    def close(self):
        if self.link:
            self.link.close()

    def _send(self, cmd, data=[]):
        if not self.link:
            packet = [0xf3, 0xfe, cmd] + data

            cr = RadioManager.open(devid=self.devid)
            cr.set_channel(self.channel)
            cr.set_data_rate(self.datarate)
            cr.set_address(self.address)
            cr.set_arc(3)

            success = False
            for i in range(50):
                res = cr.send_packet(packet)
                if res and res.ack:
                    success = True
                    break

                time.sleep(0.01)

            cr.close()

            if not success:
                raise Exception(
                    'Failed to connect to Crazyflie at {}'.format(self.uri))
        else:

            # send command (will be repeated until acked)
            pk = CRTPPacket(0xFF, [0xfe, cmd] + data)
            self.link.send_packet(pk)
            # wait up to 1s
            pk = self.link.receive_packet(0.1)
            if pk is None:
                raise Exception(
                    'Failed to connect to Crazyflie at {}'.format(self.uri))
